# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from datetime import datetime, timezone


def get_task(task_name):
    """Retrieve task object in our application instance by its fully
    qualified python name.

    Args:
        task_name (str): task's name (e.g
                         swh.loader.git.tasks.LoadDiskGitRepository)

    Returns:
        Instance of task

    """
    from swh.scheduler.celery_backend.config import app
    for module in app.conf.CELERY_IMPORTS:
        __import__(module)
    return app.tasks[task_name]


def create_task_dict(type, policy, *args, **kwargs):
    """Create a task with type and policy, scheduled for as soon as
       possible.

    Args:
        type (str): Type of oneshot task as per swh-scheduler's db
                    table task_type's column (Ex: origin-update-git,
                    swh-deposit-archive-checks)
        policy (str): oneshot or recurring policy

    Returns:
        Expected dictionary for the one-shot task scheduling api
        (swh.scheduler.backend.create_tasks)

    """
    task_extra = {}
    for extra_key in ['priority', 'retries_left']:
        if extra_key in kwargs:
            extra_val = kwargs.pop(extra_key)
            task_extra[extra_key] = extra_val

    task = {
        'policy': policy,
        'type': type,
        'next_run': datetime.now(tz=timezone.utc),
        'arguments': {
            'args': args if args else [],
            'kwargs': kwargs if kwargs else {},
        },
    }
    task.update(task_extra)
    return task


def create_oneshot_task_dict(type, *args, **kwargs):
    """Create a oneshot task scheduled for as soon as possible.

    Args:
        type (str): Type of oneshot task as per swh-scheduler's db
                    table task_type's column (Ex: origin-update-git,
                    swh-deposit-archive-checks)

    Returns:
        Expected dictionary for the one-shot task scheduling api
        (swh.scheduler.backend.create_tasks)

    """
    return create_task_dict(type, 'oneshot', *args, **kwargs)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


# Percentage of tasks with priority to schedule
PRIORITY_SLOT = 0.6

DEFAULT_CONFIG = {
    'scheduler': ('dict', {
        'cls': 'local',
        'args': {
            'db': 'dbname=softwareheritage-scheduler-dev',
        },
    })
}
# current configuration. To be set by the config loading mechanism
CONFIG = {}


def compute_nb_tasks_from(num_tasks):
    """Compute and returns the tuple, number of tasks without priority,
       number of tasks with priority.

    Args:
        num_tasks (int):

    Returns:
        tuple number of tasks without priority (int), number of tasks with
        priority (int)

    """
    if not num_tasks:
        return None, None
    return (int((1 - PRIORITY_SLOT) * num_tasks),
            int(PRIORITY_SLOT * num_tasks))


def get_scheduler(cls, args={}):
    """
    Get a scheduler object of class `scheduler_class` with arguments
    `scheduler_args`.

    Args:
        scheduler (dict): dictionary with keys:

        cls (str): scheduler's class, either 'local' or 'remote'
        args (dict): dictionary with keys, default to empty.

    Returns:
        an instance of swh.scheduler, either local or remote:

        local: swh.scheduler.backend.SchedulerBackend
        remote: swh.scheduler.api.client.RemoteScheduler

    Raises:
        ValueError if passed an unknown storage class.

    """

    if cls == 'remote':
        from .api.client import RemoteScheduler as SchedulerBackend
    elif cls == 'local':
        from .backend import SchedulerBackend
    else:
        raise ValueError('Unknown swh.scheduler class `%s`' % cls)

    return SchedulerBackend(**args)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


LISTENED_EVENTS = [
    'delete',
    'public',
    'push'
]


class SWHEvent:
    """SWH's interesting event (resulting in an origin update)

    """
    def __init__(self, evt, cnt=1):
        self.event = evt
        self.type = evt['type'].lower()
        self.url = evt['url']
        self.last_seen = evt.get('last_seen')
        self.cnt = cnt
        self.origin_type = evt.get('origin_type')

    def is_interesting(self):
        return self.type in LISTENED_EVENTS

    def get(self):
        return {
            'type': self.type,
            'url': self.url,
            'last_seen': self.last_seen,
            'cnt': self.cnt,
            'origin_type': self.origin_type
        }

    def __str__(self):
        return str(self.get())



# Copyright (C) 2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import re
import tempfile
from unittest.mock import patch

from click.testing import CliRunner
import pytest

from swh.storage.in_memory import Storage

from swh.scheduler.cli import cli
from swh.scheduler.utils import create_task_dict


CLI_CONFIG = '''
scheduler:
    cls: foo
    args: {}
'''


def invoke(scheduler, catch_exceptions, args):
    runner = CliRunner()
    with patch('swh.scheduler.cli.get_scheduler') as get_scheduler_mock, \
            tempfile.NamedTemporaryFile('a', suffix='.yml') as config_fd:
        config_fd.write(CLI_CONFIG)
        config_fd.seek(0)
        get_scheduler_mock.return_value = scheduler
        args = ['-C' + config_fd.name, '-l', 'WARNING'] + args
        result = runner.invoke(cli, args)
    if not catch_exceptions and result.exception:
        print(result.output)
        raise result.exception
    return result


def test_schedule_tasks(swh_scheduler):
    csv_data = (
        b'swh-test-ping;[["arg1", "arg2"]];{"key": "value"};'
        + datetime.datetime.utcnow().isoformat().encode() + b'\n'
        + b'swh-test-ping;[["arg3", "arg4"]];{"key": "value"};'
        + datetime.datetime.utcnow().isoformat().encode() + b'\n')
    with tempfile.NamedTemporaryFile(suffix='.csv') as csv_fd:
        csv_fd.write(csv_data)
        csv_fd.seek(0)
        result = invoke(swh_scheduler, False, [
            'task', 'schedule',
            '-d', ';',
            csv_fd.name
        ])
    expected = r'''
Created 2 tasks

Task 1
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: recurring
  Args:
    \['arg1', 'arg2'\]
  Keyword args:
    key: 'value'

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: recurring
  Args:
    \['arg3', 'arg4'\]
  Keyword args:
    key: 'value'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_schedule_tasks_columns(swh_scheduler):
    with tempfile.NamedTemporaryFile(suffix='.csv') as csv_fd:
        csv_fd.write(
            b'swh-test-ping;oneshot;["arg1", "arg2"];{"key": "value"}\n')
        csv_fd.seek(0)
        result = invoke(swh_scheduler, False, [
            'task', 'schedule',
            '-c', 'type', '-c', 'policy', '-c', 'args', '-c', 'kwargs',
            '-d', ';',
            csv_fd.name
        ])
    expected = r'''
Created 1 tasks

Task 1
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
    'arg1'
    'arg2'
  Keyword args:
    key: 'value'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_schedule_task(swh_scheduler):
    result = invoke(swh_scheduler, False, [
        'task', 'add',
        'swh-test-ping', 'arg1', 'arg2', 'key=value',
    ])
    expected = r'''
Created 1 tasks

Task 1
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: recurring
  Args:
    'arg1'
    'arg2'
  Keyword args:
    key: 'value'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks_none(swh_scheduler):
    result = invoke(swh_scheduler, False, [
        'task', 'list-pending', 'swh-test-ping',
    ])

    expected = r'''
Found 0 swh-test-ping tasks

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task2['next_run'] += datetime.timedelta(days=1)
    swh_scheduler.create_tasks([task1, task2])

    result = invoke(swh_scheduler, False, [
        'task', 'list-pending', 'swh-test-ping',
    ])

    expected = r'''
Found 1 swh-test-ping tasks

Task 1
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value1'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output

    swh_scheduler.grab_ready_tasks('swh-test-ping')

    result = invoke(swh_scheduler, False, [
        'task', 'list-pending', 'swh-test-ping',
    ])

    expected = r'''
Found 0 swh-test-ping tasks

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks_filter(swh_scheduler):
    task = create_task_dict('swh-test-multiping', 'oneshot', key='value')
    swh_scheduler.create_tasks([task])

    result = invoke(swh_scheduler, False, [
        'task', 'list-pending', 'swh-test-ping',
    ])

    expected = r'''
Found 0 swh-test-ping tasks

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks_filter_2(swh_scheduler):
    swh_scheduler.create_tasks([
        create_task_dict('swh-test-multiping', 'oneshot', key='value'),
        create_task_dict('swh-test-ping', 'oneshot', key='value2'),
    ])

    result = invoke(swh_scheduler, False, [
        'task', 'list-pending', 'swh-test-ping',
    ])

    expected = r'''
Found 1 swh-test-ping tasks

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value2'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


# Fails because "task list-pending --limit 3" only returns 2 tasks, because
# of how compute_nb_tasks_from works.
@pytest.mark.xfail
def test_list_pending_tasks_limit(swh_scheduler):
    swh_scheduler.create_tasks([
        create_task_dict('swh-test-ping', 'oneshot', key='value%d' % i)
        for i in range(10)
    ])

    result = invoke(swh_scheduler, False, [
        'task', 'list-pending', 'swh-test-ping', '--limit', '3',
    ])

    expected = r'''
Found 2 swh-test-ping tasks

Task 1
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value0'

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value1'

Task 3
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value2'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_pending_tasks_before(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task1['next_run'] += datetime.timedelta(days=3)
    task2['next_run'] += datetime.timedelta(days=1)
    swh_scheduler.create_tasks([task1, task2])

    result = invoke(swh_scheduler, False, [
        'task', 'list-pending', 'swh-test-ping', '--before',
        (datetime.date.today() + datetime.timedelta(days=2)).isoformat()
    ])

    expected = r'''
Found 1 swh-test-ping tasks

Task 2
  Next run: in a day \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Args:
  Keyword args:
    key: 'value2'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task1['next_run'] += datetime.timedelta(days=3, hours=2)
    swh_scheduler.create_tasks([task1, task2])

    swh_scheduler.grab_ready_tasks('swh-test-ping')

    result = invoke(swh_scheduler, False, [
        'task', 'list',
    ])

    expected = r'''
Found 2 tasks

Task 1
  Next run: in 3 days \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_id(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task3 = create_task_dict('swh-test-ping', 'oneshot', key='value3')
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(swh_scheduler, False, [
        'task', 'list', '--task-id', '2',
    ])

    expected = r'''
Found 1 tasks

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_id_2(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task3 = create_task_dict('swh-test-ping', 'oneshot', key='value3')
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(swh_scheduler, False, [
        'task', 'list', '--task-id', '2', '--task-id', '3'
    ])

    expected = r'''
Found 2 tasks

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

Task 3
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value3'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_type(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-multiping', 'oneshot', key='value2')
    task3 = create_task_dict('swh-test-ping', 'oneshot', key='value3')
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(swh_scheduler, False, [
        'task', 'list', '--task-type', 'swh-test-ping'
    ])

    expected = r'''
Found 2 tasks

Task 1
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

Task 3
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value3'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_limit(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task3 = create_task_dict('swh-test-ping', 'oneshot', key='value3')
    swh_scheduler.create_tasks([task1, task2, task3])

    result = invoke(swh_scheduler, False, [
        'task', 'list', '--limit', '2',
    ])

    expected = r'''
Found 2 tasks

Task 1
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_before(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task1['next_run'] += datetime.timedelta(days=3, hours=2)
    swh_scheduler.create_tasks([task1, task2])

    swh_scheduler.grab_ready_tasks('swh-test-ping')

    result = invoke(swh_scheduler, False, [
        'task', 'list', '--before',
        (datetime.date.today() + datetime.timedelta(days=2)).isoformat()
    ])

    expected = r'''
Found 1 tasks

Task 2
  Next run: just now \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value2'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def test_list_tasks_after(swh_scheduler):
    task1 = create_task_dict('swh-test-ping', 'oneshot', key='value1')
    task2 = create_task_dict('swh-test-ping', 'oneshot', key='value2')
    task1['next_run'] += datetime.timedelta(days=3, hours=2)
    swh_scheduler.create_tasks([task1, task2])

    swh_scheduler.grab_ready_tasks('swh-test-ping')

    result = invoke(swh_scheduler, False, [
        'task', 'list', '--after',
        (datetime.date.today() + datetime.timedelta(days=2)).isoformat()
    ])

    expected = r'''
Found 1 tasks

Task 1
  Next run: in 3 days \(.*\)
  Interval: 1 day, 0:00:00
  Type: swh-test-ping
  Policy: oneshot
  Status: next_run_not_scheduled
  Priority:\x20
  Args:
  Keyword args:
    key: 'value1'

'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), result.output


def _fill_storage_with_origins(storage, nb_origins):
    storage.origin_add([
        {
            'type': 'type{}'.format(i),
            'url': 'http://example.com/{}'.format(i),
        }
        for i in range(nb_origins)
    ])


@pytest.fixture
def storage():
    """An instance of swh.storage.in_memory.Storage that gets injected
    into the CLI functions."""
    storage = Storage()
    with patch('swh.scheduler.cli.get_storage') as get_storage_mock:
        get_storage_mock.return_value = storage
        yield storage


@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_task_schedule_origins_dry_run(
        swh_scheduler, storage):
    """Tests the scheduling when origin_batch_size*task_batch_size is a
    divisor of nb_origins."""
    _fill_storage_with_origins(storage, 90)

    result = invoke(swh_scheduler, False, [
        'task', 'schedule_origins', '--dry-run', 'swh-test-ping',
    ])

    # Check the output
    expected = r'''
Scheduled 3 tasks \(30 origins\).
Scheduled 6 tasks \(60 origins\).
Scheduled 9 tasks \(90 origins\).
Done.
'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), \
        repr(result.output)

    # Check scheduled tasks
    tasks = swh_scheduler.search_tasks()
    assert len(tasks) == 0


@patch('swh.scheduler.cli_utils.TASK_BATCH_SIZE', 3)
def test_task_schedule_origins(swh_scheduler, storage):
    """Tests the scheduling when neither origin_batch_size or
    task_batch_size is a divisor of nb_origins."""
    _fill_storage_with_origins(storage, 70)

    result = invoke(swh_scheduler, False, [
        'task', 'schedule_origins', 'swh-test-ping',
        '--batch-size', '20',
    ])

    # Check the output
    expected = r'''
Scheduled 3 tasks \(60 origins\).
Scheduled 4 tasks \(70 origins\).
Done.
'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), \
        repr(result.output)

    # Check scheduled tasks
    tasks = swh_scheduler.search_tasks()
    assert len(tasks) == 4
    assert tasks[0]['arguments']['args'] == [list(range(1, 21))]
    assert tasks[1]['arguments']['args'] == [list(range(21, 41))]
    assert tasks[2]['arguments']['args'] == [list(range(41, 61))]
    assert tasks[3]['arguments']['args'] == [list(range(61, 71))]
    assert all(task['arguments']['kwargs'] == {} for task in tasks)


def test_task_schedule_origins_kwargs(swh_scheduler, storage):
    """Tests support of extra keyword-arguments."""
    _fill_storage_with_origins(storage, 30)

    result = invoke(swh_scheduler, False, [
        'task', 'schedule_origins', 'swh-test-ping',
        '--batch-size', '20',
        'key1="value1"', 'key2="value2"',
    ])

    # Check the output
    expected = r'''
Scheduled 2 tasks \(30 origins\).
Done.
'''.lstrip()
    assert result.exit_code == 0, result.output
    assert re.fullmatch(expected, result.output, re.MULTILINE), \
        repr(result.output)

    # Check scheduled tasks
    tasks = swh_scheduler.search_tasks()
    assert len(tasks) == 2
    assert tasks[0]['arguments']['args'] == [list(range(1, 21))]
    assert tasks[1]['arguments']['args'] == [list(range(21, 31))]
    assert all(task['arguments']['kwargs'] ==
               {'key1': 'value1', 'key2': 'value2'}
               for task in tasks)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from datetime import timezone
from unittest.mock import patch

from swh.scheduler import utils


class UtilsTest(unittest.TestCase):

    @patch('swh.scheduler.utils.datetime')
    def test_create_oneshot_task_dict_simple(self, mock_datetime):
        mock_datetime.now.return_value = 'some-date'

        actual_task = utils.create_oneshot_task_dict('some-task-type')

        expected_task = {
            'policy': 'oneshot',
            'type': 'some-task-type',
            'next_run': 'some-date',
            'arguments': {
                'args': [],
                'kwargs': {},
            },
        }

        self.assertEqual(actual_task, expected_task)
        mock_datetime.now.assert_called_once_with(tz=timezone.utc)

    @patch('swh.scheduler.utils.datetime')
    def test_create_oneshot_task_dict_other_call(self, mock_datetime):
        mock_datetime.now.return_value = 'some-other-date'

        actual_task = utils.create_oneshot_task_dict(
            'some-task-type', 'arg0', 'arg1',
            priority='high', other_stuff='normal'
        )

        expected_task = {
            'policy': 'oneshot',
            'type': 'some-task-type',
            'next_run': 'some-other-date',
            'arguments': {
                'args': ('arg0', 'arg1'),
                'kwargs': {'other_stuff': 'normal'},
            },
            'priority': 'high',
        }

        self.assertEqual(actual_task, expected_task)
        mock_datetime.now.assert_called_once_with(tz=timezone.utc)

    @patch('swh.scheduler.utils.datetime')
    def test_create_task_dict(self, mock_datetime):
        mock_datetime.now.return_value = 'date'

        actual_task = utils.create_task_dict(
            'task-type', 'recurring', 'arg0', 'arg1',
            priority='low', other_stuff='normal', retries_left=3
        )

        expected_task = {
            'policy': 'recurring',
            'type': 'task-type',
            'next_run': 'date',
            'arguments': {
                'args': ('arg0', 'arg1'),
                'kwargs': {'other_stuff': 'normal'},
            },
            'priority': 'low',
            'retries_left': 3,
        }

        self.assertEqual(actual_task, expected_task)
        mock_datetime.now.assert_called_once_with(tz=timezone.utc)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest

from hypothesis import given
from hypothesis.strategies import sampled_from, text

from swh.scheduler.updater.events import LISTENED_EVENTS, SWHEvent
from swh.scheduler.updater.ghtorrent import events

from . import UpdaterTestUtil


def event_values_ko():
    return set(events['evt']).union(
        set(events['ent'])).difference(
        set(LISTENED_EVENTS))


WRONG_EVENTS = sorted(list(event_values_ko()))


class EventTest(UpdaterTestUtil, unittest.TestCase):
    @given(sampled_from(LISTENED_EVENTS), text(), text())
    def test_is_interesting_ok(self, event_type, name, origin_type):
        evt = self._make_simple_event(event_type, name, origin_type)
        self.assertTrue(SWHEvent(evt).is_interesting())

    @given(text(), text(), text())
    def test_is_interested_with_noisy_event_should_be_ko(
            self, event_type, name, origin_type):
        if event_type in LISTENED_EVENTS:
            # just in case something good is generated, skip it
            return
        evt = self._make_simple_event(event_type, name, origin_type)
        self.assertFalse(SWHEvent(evt).is_interesting())

    @given(sampled_from(WRONG_EVENTS), text(), text())
    def test_is_interesting_ko(self, event_type, name, origin_type):
        evt = self._make_simple_event(event_type, name, origin_type)
        self.assertFalse(SWHEvent(evt).is_interesting())



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


"""Elastic Search backend

"""
from copy import deepcopy

from swh.core import utils
from elasticsearch import Elasticsearch
from elasticsearch import helpers


DEFAULT_CONFIG = {
    'elastic_search': {
        'storage_nodes': {'host': 'localhost', 'port': 9200},
        'index_name_prefix': 'swh-tasks',
        'client_options': {
            'sniff_on_start': False,
            'sniff_on_connection_fail': True,
            'http_compress': False,
        },
    },
}


class SWHElasticSearchClient:
    def __init__(self, **config):
        self.config = deepcopy(DEFAULT_CONFIG)
        self.config.update(config)
        es_conf = self.config['elastic_search']
        options = es_conf.get('client_options', {})
        self.storage = Elasticsearch(
            # nodes to use by default
            es_conf['storage_nodes'],
            # auto detect cluster's status
            sniff_on_start=options['sniff_on_start'],
            sniff_on_connection_fail=options['sniff_on_connection_fail'],
            sniffer_timeout=60,
            # compression or not
            http_compress=options['http_compress'])
        self.index_name_prefix = es_conf['index_name_prefix']
        # document's index type (cf. ../../data/elastic-template.json)
        self.doc_type = 'task'

    def compute_index_name(self, year, month):
        """Given a year, month, compute the index's name.

        """
        return '%s-%s-%s' % (
            self.index_name_prefix, year, '%02d' % month)

    def index(self, data):
        """Index given data to elasticsearch.

        The field 'ended' in data is used to compute the index to
        index data to.

        """
        date = data['ended']
        index_name = self.compute_index_name(date.year, date.month)
        return self.storage.index(index=index_name,
                                  doc_type=self.doc_type,
                                  body=data)

    def mget(self, index_name, doc_ids, chunk_size=500,
             source=True, log=None):
        """Retrieve document's full content according to their ids as per
           source's setup.

           The `source` permits to retrieve only what's of interest to
           us, e.g:
           - source=True ; gives back the original indexed data
           - source=False ; returns without the original _source field
           - source=['task_id'] ; returns only task_id in the _source field

        Args:
            index_name (str): Name of the concerned index.
            doc_ids (generator): Generator of ids to retrieve
            chunk_size (int): Number of documents chunk to send for retrieval
            source (bool/[str]): Source of information to return

        Yields:
            document indexed as per source's setup

        """
        if isinstance(source, list):
            source = {'_source': ','.join(source)}
        else:
            source = {'_source': str(source).lower()}

        for ids in utils.grouper(doc_ids, n=1000):
            res = self.storage.mget(body={'ids': list(ids)},
                                    index=index_name,
                                    doc_type=self.doc_type,
                                    params=source)
            if not res:
                if log:
                    log.error('Error during retrieval of data, skipping!')
                continue

            for doc in res['docs']:
                found = doc.get('found')
                if not found:
                    msg = 'Doc id %s not found, not indexed yet' % doc['_id']
                    if log:
                        log.warning(msg)
                    continue
                yield doc['_source']

    def _streaming_bulk(self, index_name, doc_stream, chunk_size=500,
                        log=None):
        """Bulk index data and returns the successful indexed data's
           identifier.

        Args:
            index_name (str): Name of the concerned index.
            doc_stream (generator): Generator of documents to index
            chunk_size (int): Number of documents chunk to send for indexation

        Yields:
            document id indexed

        """
        actions = ({'_index': index_name,
                    '_op_type': 'index',
                    '_type': self.doc_type,
                    '_source': data} for data in doc_stream)
        for ok, result in helpers.streaming_bulk(client=self.storage,
                                                 actions=actions,
                                                 chunk_size=chunk_size,
                                                 raise_on_error=False,
                                                 raise_on_exception=False):
            if not ok:
                if log:
                    log.error('Error during %s indexation. Skipping.' % result)
                continue
            yield result['index']['_id']

    def streaming_bulk(self, index_name, doc_stream, chunk_size=500,
                       source=True, log=None):
        """Bulk index data and returns the successful indexed data as per
           source's setup.

           the `source` permits to retrieve only what's of interest to
           us, e.g:

           - source=True ; gives back the original indexed data
           - source=False ; returns without the original _source field
           - source=['task_id'] ; returns only task_id in the _source field

        Args:
            index_name (str): Name of the concerned index.
            doc_stream (generator): Document generator to index
            chunk_size (int): Number of documents chunk to send
            source (bool, [str]): the information to return

        """

        indexed_ids = self._streaming_bulk(
            index_name, doc_stream, chunk_size=chunk_size, log=log)
        yield from self.mget(index_name, indexed_ids, chunk_size=chunk_size,
                             source=source, log=log)



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import copy
import datetime
import os
import random
import unittest
import uuid
from collections import defaultdict

import psycopg2
from arrow import utcnow
import pytest

from swh.core.tests.db_testing import SingleDbTestFixture
from swh.scheduler import get_scheduler

from . import SQL_DIR


TASK_TYPES = {
    'git': {
        'type': 'update-git',
        'description': 'Update a git repository',
        'backend_name': 'swh.loader.git.tasks.UpdateGitRepository',
        'default_interval': datetime.timedelta(days=64),
        'min_interval': datetime.timedelta(hours=12),
        'max_interval': datetime.timedelta(days=64),
        'backoff_factor': 2,
        'max_queue_length': None,
        'num_retries': 7,
        'retry_delay': datetime.timedelta(hours=2),
    },
    'hg': {
        'type': 'update-hg',
        'description': 'Update a mercurial repository',
        'backend_name': 'swh.loader.mercurial.tasks.UpdateHgRepository',
        'default_interval': datetime.timedelta(days=64),
        'min_interval': datetime.timedelta(hours=12),
        'max_interval': datetime.timedelta(days=64),
        'backoff_factor': 2,
        'max_queue_length': None,
        'num_retries': 7,
        'retry_delay': datetime.timedelta(hours=2),
    },
}

TEMPLATES = {
    'git': {
        'type': 'update-git',
        'arguments': {
            'args': [],
            'kwargs': {},
        },
        'next_run': None,
    },
    'hg': {
        'type': 'update-hg',
        'arguments': {
            'args': [],
            'kwargs': {},
        },
        'next_run': None,
        'policy': 'oneshot',
    }
}


def subdict(d, keys=None, excl=()):
    if keys is None:
        keys = [k for k in d.keys()]
    return {k: d[k] for k in keys if k not in excl}


@pytest.mark.db
class CommonSchedulerTest(SingleDbTestFixture):
    TEST_DB_NAME = 'softwareheritage-scheduler-test'
    TEST_DB_DUMP = os.path.join(SQL_DIR, '*.sql')

    def tearDown(self):
        self.empty_tables()
        super().tearDown()

    def empty_tables(self, whitelist=["priority_ratio"]):
        query = """SELECT table_name FROM information_schema.tables
                   WHERE table_schema = %%s and
                         table_name not in (%s)
                """ % ','.join(map(lambda t: "'%s'" % t, whitelist))
        self.cursor.execute(query, ('public', ))

        tables = set(table for (table,) in self.cursor.fetchall())

        for table in tables:
            self.cursor.execute('truncate table %s cascade' % table)
        self.conn.commit()

    def test_add_task_type(self):
        tt = TASK_TYPES['git']
        self.backend.create_task_type(tt)
        self.assertEqual(tt, self.backend.get_task_type(tt['type']))
        with self.assertRaisesRegex(psycopg2.IntegrityError,
                                    r'\(type\)=\(%s\)' % tt['type']):
            self.backend.create_task_type(tt)

        tt2 = TASK_TYPES['hg']
        self.backend.create_task_type(tt2)
        self.assertEqual(tt, self.backend.get_task_type(tt['type']))
        self.assertEqual(tt2, self.backend.get_task_type(tt2['type']))

    def test_get_task_types(self):
        tt, tt2 = TASK_TYPES['git'], TASK_TYPES['hg']
        self.backend.create_task_type(tt)
        self.backend.create_task_type(tt2)
        self.assertCountEqual([tt2, tt], self.backend.get_task_types())

    @staticmethod
    def _task_from_template(template, next_run, priority, *args, **kwargs):
        ret = copy.deepcopy(template)
        ret['next_run'] = next_run
        if priority:
            ret['priority'] = priority
        if args:
            ret['arguments']['args'] = list(args)
        if kwargs:
            ret['arguments']['kwargs'] = kwargs
        return ret

    def _pop_priority(self, priorities):
        if not priorities:
            return None
        for priority, remains in priorities.items():
            if remains > 0:
                priorities[priority] = remains - 1
                return priority
        return None

    def _tasks_from_template(self, template, max_timestamp, num,
                             num_priority=0, priorities=None):
        if num_priority and priorities:
            priorities = {
                priority: ratio * num_priority
                for priority, ratio in priorities.items()
            }

        tasks = []
        for i in range(num + num_priority):
            priority = self._pop_priority(priorities)
            tasks.append(self._task_from_template(
                template,
                max_timestamp - datetime.timedelta(microseconds=i),
                priority,
                'argument-%03d' % i,
                **{'kwarg%03d' % i: 'bogus-kwarg'}
            ))
        return tasks

    def _create_task_types(self):
        for tt in TASK_TYPES.values():
            self.backend.create_task_type(tt)

    def test_create_tasks(self):
        priority_ratio = self._priority_ratio()
        self._create_task_types()
        num_tasks_priority = 100
        tasks_1 = self._tasks_from_template(
            TEMPLATES['git'], utcnow(), 100)
        tasks_2 = self._tasks_from_template(
            TEMPLATES['hg'], utcnow(), 100,
            num_tasks_priority, priorities=priority_ratio)
        tasks = tasks_1 + tasks_2

        # tasks are returned only once with their ids
        ret1 = self.backend.create_tasks(tasks + tasks_1 + tasks_2)
        set_ret1 = set([t['id'] for t in ret1])

        # creating the same set result in the same ids
        ret = self.backend.create_tasks(tasks)
        set_ret = set([t['id'] for t in ret])

        # Idempotence results
        self.assertEqual(set_ret, set_ret1)
        self.assertEqual(len(ret), len(ret1))

        ids = set()
        actual_priorities = defaultdict(int)

        for task, orig_task in zip(ret, tasks):
            task = copy.deepcopy(task)
            task_type = TASK_TYPES[orig_task['type'].split('-')[-1]]
            self.assertNotIn(task['id'], ids)
            self.assertEqual(task['status'], 'next_run_not_scheduled')
            self.assertEqual(task['current_interval'],
                             task_type['default_interval'])
            self.assertEqual(task['policy'], orig_task.get('policy',
                                                           'recurring'))
            priority = task.get('priority')
            if priority:
                actual_priorities[priority] += 1

            self.assertEqual(task['retries_left'],
                             task_type['num_retries'] or 0)
            ids.add(task['id'])
            del task['id']
            del task['status']
            del task['current_interval']
            del task['retries_left']
            if 'policy' not in orig_task:
                del task['policy']
            if 'priority' not in orig_task:
                del task['priority']
                self.assertEqual(task, orig_task)

        self.assertEqual(dict(actual_priorities), {
            priority: int(ratio * num_tasks_priority)
            for priority, ratio in priority_ratio.items()
        })

    def test_peek_ready_tasks_no_priority(self):
        self._create_task_types()
        t = utcnow()
        task_type = TEMPLATES['git']['type']
        tasks = self._tasks_from_template(TEMPLATES['git'], t, 100)
        random.shuffle(tasks)
        self.backend.create_tasks(tasks)

        ready_tasks = self.backend.peek_ready_tasks(task_type)
        self.assertEqual(len(ready_tasks), len(tasks))
        for i in range(len(ready_tasks) - 1):
            self.assertLessEqual(ready_tasks[i]['next_run'],
                                 ready_tasks[i+1]['next_run'])

        # Only get the first few ready tasks
        limit = random.randrange(5, 5 + len(tasks)//2)
        ready_tasks_limited = self.backend.peek_ready_tasks(
            task_type, num_tasks=limit)

        self.assertEqual(len(ready_tasks_limited), limit)
        self.assertCountEqual(ready_tasks_limited, ready_tasks[:limit])

        # Limit by timestamp
        max_ts = tasks[limit-1]['next_run']
        ready_tasks_timestamped = self.backend.peek_ready_tasks(
            task_type, timestamp=max_ts)

        for ready_task in ready_tasks_timestamped:
            self.assertLessEqual(ready_task['next_run'], max_ts)

        # Make sure we get proper behavior for the first ready tasks
        self.assertCountEqual(
            ready_tasks[:len(ready_tasks_timestamped)],
            ready_tasks_timestamped,
        )

        # Limit by both
        ready_tasks_both = self.backend.peek_ready_tasks(
            task_type, timestamp=max_ts, num_tasks=limit//3)
        self.assertLessEqual(len(ready_tasks_both), limit//3)
        for ready_task in ready_tasks_both:
            self.assertLessEqual(ready_task['next_run'], max_ts)
            self.assertIn(ready_task, ready_tasks[:limit//3])

    def _priority_ratio(self):
        self.cursor.execute('select id, ratio from priority_ratio')
        priority_ratio = {}
        for row in self.cursor.fetchall():
            priority_ratio[row[0]] = row[1]
        return priority_ratio

    def test_peek_ready_tasks_mixed_priorities(self):
        priority_ratio = self._priority_ratio()
        self._create_task_types()
        t = utcnow()
        task_type = TEMPLATES['git']['type']
        num_tasks_priority = 100
        num_tasks_no_priority = 100
        # Create tasks with and without priorities
        tasks = self._tasks_from_template(
            TEMPLATES['git'], t,
            num=num_tasks_no_priority,
            num_priority=num_tasks_priority,
            priorities=priority_ratio)

        random.shuffle(tasks)
        self.backend.create_tasks(tasks)

        # take all available tasks
        ready_tasks = self.backend.peek_ready_tasks(
            task_type)

        self.assertEqual(len(ready_tasks), len(tasks))
        self.assertEqual(num_tasks_priority + num_tasks_no_priority,
                         len(ready_tasks))

        count_tasks_per_priority = defaultdict(int)
        for task in ready_tasks:
            priority = task.get('priority')
            if priority:
                count_tasks_per_priority[priority] += 1

        self.assertEqual(dict(count_tasks_per_priority), {
            priority: int(ratio * num_tasks_priority)
            for priority, ratio in priority_ratio.items()
        })

        # Only get some ready tasks
        num_tasks = random.randrange(5, 5 + num_tasks_no_priority//2)
        num_tasks_priority = random.randrange(5, num_tasks_priority//2)
        ready_tasks_limited = self.backend.peek_ready_tasks(
            task_type, num_tasks=num_tasks,
            num_tasks_priority=num_tasks_priority)

        count_tasks_per_priority = defaultdict(int)
        for task in ready_tasks_limited:
            priority = task.get('priority')
            count_tasks_per_priority[priority] += 1

        import math
        for priority, ratio in priority_ratio.items():
            expected_count = math.ceil(ratio * num_tasks_priority)
            actual_prio = count_tasks_per_priority[priority]
            self.assertTrue(
                actual_prio == expected_count or
                actual_prio == expected_count + 1)

        self.assertEqual(count_tasks_per_priority[None], num_tasks)

    def test_grab_ready_tasks(self):
        priority_ratio = self._priority_ratio()
        self._create_task_types()
        t = utcnow()
        task_type = TEMPLATES['git']['type']
        num_tasks_priority = 100
        num_tasks_no_priority = 100
        # Create tasks with and without priorities
        tasks = self._tasks_from_template(
            TEMPLATES['git'], t,
            num=num_tasks_no_priority,
            num_priority=num_tasks_priority,
            priorities=priority_ratio)
        random.shuffle(tasks)
        self.backend.create_tasks(tasks)

        first_ready_tasks = self.backend.peek_ready_tasks(
            task_type, num_tasks=10, num_tasks_priority=10)
        grabbed_tasks = self.backend.grab_ready_tasks(
            task_type, num_tasks=10, num_tasks_priority=10)

        for peeked, grabbed in zip(first_ready_tasks, grabbed_tasks):
            self.assertEqual(peeked['status'], 'next_run_not_scheduled')
            del peeked['status']
            self.assertEqual(grabbed['status'], 'next_run_scheduled')
            del grabbed['status']
            self.assertEqual(peeked, grabbed)
            self.assertEqual(peeked['priority'], grabbed['priority'])

    def test_get_tasks(self):
        self._create_task_types()
        t = utcnow()
        tasks = self._tasks_from_template(TEMPLATES['git'], t, 100)
        tasks = self.backend.create_tasks(tasks)
        random.shuffle(tasks)
        while len(tasks) > 1:
            length = random.randrange(1, len(tasks))
            cur_tasks = tasks[:length]
            tasks[:length] = []

            ret = self.backend.get_tasks(task['id'] for task in cur_tasks)
            self.assertCountEqual(ret, cur_tasks)

    def test_search_tasks(self):
        self._create_task_types()
        t = utcnow()
        tasks = self._tasks_from_template(TEMPLATES['git'], t, 100)
        tasks = self.backend.create_tasks(tasks)
        self.assertCountEqual(self.backend.search_tasks(), tasks)

    def test_filter_task_to_archive(self):
        """Filtering only list disabled recurring or completed oneshot tasks

        """
        self._create_task_types()
        _time = utcnow()
        recurring = self._tasks_from_template(TEMPLATES['git'], _time, 12)
        oneshots = self._tasks_from_template(TEMPLATES['hg'], _time, 12)
        total_tasks = len(recurring) + len(oneshots)

        # simulate scheduling tasks
        pending_tasks = self.backend.create_tasks(recurring + oneshots)
        backend_tasks = [{
            'task': task['id'],
            'backend_id': str(uuid.uuid4()),
            'scheduled': utcnow(),
        } for task in pending_tasks]
        self.backend.mass_schedule_task_runs(backend_tasks)

        # we simulate the task are being done
        _tasks = []
        for task in backend_tasks:
            t = self.backend.end_task_run(
                task['backend_id'], status='eventful')
            _tasks.append(t)

        # Randomly update task's status per policy
        status_per_policy = {'recurring': 0, 'oneshot': 0}
        status_choice = {
            # policy: [tuple (1-for-filtering, 'associated-status')]
            'recurring': [(1, 'disabled'),
                          (0, 'completed'),
                          (0, 'next_run_not_scheduled')],
            'oneshot': [(0, 'next_run_not_scheduled'),
                        (1, 'disabled'),
                        (1, 'completed')]
        }

        tasks_to_update = defaultdict(list)
        _task_ids = defaultdict(list)
        # randomize 'disabling' recurring task or 'complete' oneshot task
        for task in pending_tasks:
            policy = task['policy']
            _task_ids[policy].append(task['id'])
            status = random.choice(status_choice[policy])
            if status[0] != 1:
                continue
            # elected for filtering
            status_per_policy[policy] += status[0]
            tasks_to_update[policy].append(task['id'])

        self.backend.disable_tasks(tasks_to_update['recurring'])
        # hack: change the status to something else than completed/disabled
        self.backend.set_status_tasks(
            _task_ids['oneshot'], status='next_run_not_scheduled')
        # complete the tasks to update
        self.backend.set_status_tasks(
            tasks_to_update['oneshot'], status='completed')

        total_tasks_filtered = (status_per_policy['recurring'] +
                                status_per_policy['oneshot'])

        # retrieve tasks to archive
        after = _time.shift(days=-1).format('YYYY-MM-DD')
        before = utcnow().shift(days=1).format('YYYY-MM-DD')
        tasks_to_archive = list(self.backend.filter_task_to_archive(
            after_ts=after, before_ts=before, limit=total_tasks))

        self.assertEqual(len(tasks_to_archive), total_tasks_filtered)

        actual_filtered_per_status = {'recurring': 0, 'oneshot': 0}
        for task in tasks_to_archive:
            actual_filtered_per_status[task['task_policy']] += 1

        self.assertEqual(actual_filtered_per_status, status_per_policy)

    def test_delete_archived_tasks(self):
        self._create_task_types()
        _time = utcnow()
        recurring = self._tasks_from_template(
            TEMPLATES['git'], _time, 12)
        oneshots = self._tasks_from_template(
            TEMPLATES['hg'], _time, 12)
        total_tasks = len(recurring) + len(oneshots)
        pending_tasks = self.backend.create_tasks(recurring + oneshots)
        backend_tasks = [{
            'task': task['id'],
            'backend_id': str(uuid.uuid4()),
            'scheduled': utcnow(),
        } for task in pending_tasks]
        self.backend.mass_schedule_task_runs(backend_tasks)

        _tasks = []
        percent = random.randint(0, 100)  # random election removal boundary
        for task in backend_tasks:
            t = self.backend.end_task_run(
                task['backend_id'], status='eventful')
            c = random.randint(0, 100)
            if c <= percent:
                _tasks.append({'task_id': t['task'], 'task_run_id': t['id']})

        self.backend.delete_archived_tasks(_tasks)

        self.cursor.execute('select count(*) from task')
        tasks_count = self.cursor.fetchone()

        self.cursor.execute('select count(*) from task_run')
        tasks_run_count = self.cursor.fetchone()

        self.assertEqual(tasks_count[0], total_tasks - len(_tasks))
        self.assertEqual(tasks_run_count[0], total_tasks - len(_tasks))

    def test_get_task_runs_no_task(self):
        '''No task exist in the scheduler's db, get_task_runs() should always return an
        empty list.

        '''
        self.assertFalse(self.backend.get_task_runs(task_ids=()))
        self.assertFalse(self.backend.get_task_runs(task_ids=(1, 2, 3)))
        self.assertFalse(self.backend.get_task_runs(task_ids=(1, 2, 3),
                                                    limit=10))

    def test_get_task_runs_no_task_executed(self):
        '''No task has been executed yet, get_task_runs() should always return an empty
        list.

        '''
        self._create_task_types()
        _time = utcnow()
        recurring = self._tasks_from_template(
            TEMPLATES['git'], _time, 12)
        oneshots = self._tasks_from_template(
            TEMPLATES['hg'], _time, 12)
        self.backend.create_tasks(recurring + oneshots)

        self.assertFalse(self.backend.get_task_runs(
            task_ids=()))
        self.assertFalse(self.backend.get_task_runs(
            task_ids=(1, 2, 3)))
        self.assertFalse(self.backend.get_task_runs(
            task_ids=(1, 2, 3), limit=10))

    def test_get_task_runs_with_scheduled(self):
        '''Some tasks have been scheduled but not executed yet, get_task_runs() should
        not return an empty list. limit should behave as expected.

        '''
        self._create_task_types()
        _time = utcnow()
        recurring = self._tasks_from_template(
            TEMPLATES['git'], _time, 12)
        oneshots = self._tasks_from_template(
            TEMPLATES['hg'], _time, 12)
        total_tasks = len(recurring) + len(oneshots)
        pending_tasks = self.backend.create_tasks(recurring + oneshots)
        backend_tasks = [{
            'task': task['id'],
            'backend_id': str(uuid.uuid4()),
            'scheduled': utcnow(),
        } for task in pending_tasks]
        self.backend.mass_schedule_task_runs(backend_tasks)

        self.assertFalse(self.backend.get_task_runs(
            task_ids=[total_tasks + 1]))

        btask = backend_tasks[0]
        runs = self.backend.get_task_runs(
            task_ids=[btask['task']])
        self.assertEqual(len(runs), 1)
        run = runs[0]

        self.assertEqual(subdict(run, excl=('id',)),
                         {'task': btask['task'],
                          'backend_id': btask['backend_id'],
                          'scheduled': btask['scheduled'],
                          'started': None,
                          'ended': None,
                          'metadata': None,
                          'status': 'scheduled',
                          })

        runs = self.backend.get_task_runs(
            task_ids=[bt['task'] for bt in backend_tasks], limit=2)
        self.assertEqual(len(runs), 2)

        runs = self.backend.get_task_runs(
            task_ids=[bt['task'] for bt in backend_tasks])
        self.assertEqual(len(runs), total_tasks)

        keys = ('task', 'backend_id', 'scheduled')
        self.assertEqual(sorted([subdict(x, keys) for x in runs],
                                key=lambda x: x['task']),
                         backend_tasks)

    def test_get_task_runs_with_executed(self):
        '''Some tasks have been executed, get_task_runs() should
        not return an empty list. limit should behave as expected.

        '''
        self._create_task_types()
        _time = utcnow()
        recurring = self._tasks_from_template(
            TEMPLATES['git'], _time, 12)
        oneshots = self._tasks_from_template(
            TEMPLATES['hg'], _time, 12)
        pending_tasks = self.backend.create_tasks(recurring + oneshots)
        backend_tasks = [{
            'task': task['id'],
            'backend_id': str(uuid.uuid4()),
            'scheduled': utcnow(),
        } for task in pending_tasks]
        self.backend.mass_schedule_task_runs(backend_tasks)

        btask = backend_tasks[0]
        ts = utcnow()
        self.backend.start_task_run(btask['backend_id'],
                                    metadata={'something': 'stupid'},
                                    timestamp=ts)
        runs = self.backend.get_task_runs(task_ids=[btask['task']])
        self.assertEqual(len(runs), 1)
        self.assertEqual(subdict(runs[0], excl=('id')), {
            'task': btask['task'],
            'backend_id': btask['backend_id'],
            'scheduled': btask['scheduled'],
            'started': ts,
            'ended': None,
            'metadata': {'something': 'stupid'},
            'status': 'started',
            })

        ts2 = utcnow()
        self.backend.end_task_run(btask['backend_id'],
                                  metadata={'other': 'stuff'},
                                  timestamp=ts2,
                                  status='eventful')
        runs = self.backend.get_task_runs(task_ids=[btask['task']])
        self.assertEqual(len(runs), 1)
        self.assertEqual(subdict(runs[0], excl=('id')), {
            'task': btask['task'],
            'backend_id': btask['backend_id'],
            'scheduled': btask['scheduled'],
            'started': ts,
            'ended': ts2,
            'metadata': {'something': 'stupid', 'other': 'stuff'},
            'status': 'eventful',
            })


class LocalSchedulerTest(CommonSchedulerTest, unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.config = {'db': 'dbname=' + self.TEST_DB_NAME}
        self.backend = get_scheduler('local', self.config)



# Copyright (C) 2016-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import arrow
import click
import csv
import itertools
import json
import locale
import logging
import time
import datetime

from swh.core import utils, config
from swh.storage import get_storage
from swh.storage.algos.origin import iter_origins

from . import compute_nb_tasks_from
from .backend_es import SWHElasticSearchClient
from . import get_scheduler, DEFAULT_CONFIG
from .cli_utils import parse_options, schedule_origin_batches


locale.setlocale(locale.LC_ALL, '')
ARROW_LOCALE = locale.getlocale(locale.LC_TIME)[0]


class DateTimeType(click.ParamType):
    name = 'time and date'

    def convert(self, value, param, ctx):
        if not isinstance(value, arrow.Arrow):
            value = arrow.get(value)

        return value


DATETIME = DateTimeType()
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


def format_dict(d):
    ret = {}
    for k, v in d.items():
        if isinstance(v, (arrow.Arrow, datetime.date, datetime.datetime)):
            v = arrow.get(v).format()
        elif isinstance(v, dict):
            v = format_dict(v)
        ret[k] = v
    return ret


def pretty_print_list(list, indent=0):
    """Pretty-print a list"""
    return ''.join('%s%r\n' % (' ' * indent, item) for item in list)


def pretty_print_dict(dict, indent=0):
    """Pretty-print a list"""
    return ''.join('%s%s: %r\n' %
                   (' ' * indent, click.style(key, bold=True), value)
                   for key, value in sorted(dict.items()))


def pretty_print_run(run, indent=4):
    fmt = ('{indent}{backend_id} [{status}]\n'
           '{indent}  scheduled: {scheduled} [{started}:{ended}]')
    return fmt.format(indent=' '*indent, **format_dict(run))


def pretty_print_task(task, full=False):
    """Pretty-print a task

    If 'full' is True, also print the status and priority fields.

    >>> task = {
    ...     'id': 1234,
    ...     'arguments': {
    ...         'args': ['foo', 'bar', True],
    ...         'kwargs': {'key': 'value', 'key2': 42},
    ...     },
    ...     'current_interval': datetime.timedelta(hours=1),
    ...     'next_run': datetime.datetime(2019, 2, 21, 13, 52, 35, 407818),
    ...     'policy': 'oneshot',
    ...     'priority': None,
    ...     'status': 'next_run_not_scheduled',
    ...     'type': 'test_task',
    ... }
    >>> print(click.unstyle(pretty_print_task(task)))
    Task 1234
      Next run: ... (2019-02-21 13:52:35+00:00)
      Interval: 1:00:00
      Type: test_task
      Policy: oneshot
      Args:
        'foo'
        'bar'
        True
      Keyword args:
        key: 'value'
        key2: 42
    <BLANKLINE>
    >>> print(click.unstyle(pretty_print_task(task, full=True)))
    Task 1234
      Next run: ... (2019-02-21 13:52:35+00:00)
      Interval: 1:00:00
      Type: test_task
      Policy: oneshot
      Status: next_run_not_scheduled
      Priority:\x20
      Args:
        'foo'
        'bar'
        True
      Keyword args:
        key: 'value'
        key2: 42
    <BLANKLINE>
    """
    next_run = arrow.get(task['next_run'])
    lines = [
        '%s %s\n' % (click.style('Task', bold=True), task['id']),
        click.style('  Next run: ', bold=True),
        "%s (%s)" % (next_run.humanize(locale=ARROW_LOCALE),
                     next_run.format()),
        '\n',
        click.style('  Interval: ', bold=True),
        str(task['current_interval']), '\n',
        click.style('  Type: ', bold=True), task['type'] or '', '\n',
        click.style('  Policy: ', bold=True), task['policy'] or '', '\n',
        ]
    if full:
        lines += [
            click.style('  Status: ', bold=True),
            task['status'] or '', '\n',
            click.style('  Priority: ', bold=True),
            task['priority'] or '', '\n',
        ]
    lines += [
        click.style('  Args:\n', bold=True),
        pretty_print_list(task['arguments']['args'], indent=4),
        click.style('  Keyword args:\n', bold=True),
        pretty_print_dict(task['arguments']['kwargs'], indent=4),
    ]

    return ''.join(lines)


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('--config-file', '-C', default=None,
              type=click.Path(exists=True, dir_okay=False,),
              help="Configuration file.")
@click.option('--database', '-d', default=None,
              help="Scheduling database DSN (imply cls is 'local')")
@click.option('--url', '-u', default=None,
              help="Scheduler's url access (imply cls is 'remote')")
@click.option('--log-level', '-l', default='INFO',
              type=click.Choice(logging._nameToLevel.keys()),
              help="Log level (default to INFO)")
@click.option('--no-stdout', is_flag=True, default=False,
              help="Do NOT output logs on the console")
@click.pass_context
def cli(ctx, config_file, database, url, log_level, no_stdout):
    """Software Heritage Scheduler CLI interface

    Default to use the the local scheduler instance (plugged to the
    main scheduler db).

    """
    from swh.scheduler.celery_backend.config import setup_log_handler
    log_level = setup_log_handler(
        loglevel=log_level, colorize=False,
        format='[%(levelname)s] %(name)s -- %(message)s',
        log_console=not no_stdout)

    ctx.ensure_object(dict)

    logger = logging.getLogger(__name__)
    scheduler = None
    conf = config.read(config_file, DEFAULT_CONFIG)
    if 'scheduler' not in conf:
        raise ValueError("missing 'scheduler' configuration")

    if database:
        conf['scheduler']['cls'] = 'local'
        conf['scheduler']['args']['db'] = database
    elif url:
        conf['scheduler']['cls'] = 'remote'
        conf['scheduler']['args'] = {'url': url}
    sched_conf = conf['scheduler']
    try:
        logger.debug('Instanciating scheduler with %s' % (
            sched_conf))
        scheduler = get_scheduler(**sched_conf)
    except ValueError:
        # it's the subcommand to decide whether not having a proper
        # scheduler instance is a problem.
        pass

    ctx.obj['scheduler'] = scheduler
    ctx.obj['config'] = conf
    ctx.obj['loglevel'] = log_level


@cli.group('task')
@click.pass_context
def task(ctx):
    """Manipulate tasks."""
    pass


@task.command('schedule')
@click.option('--columns', '-c', multiple=True,
              default=['type', 'args', 'kwargs', 'next_run'],
              type=click.Choice([
                  'type', 'args', 'kwargs', 'policy', 'next_run']),
              help='columns present in the CSV file')
@click.option('--delimiter', '-d', default=',')
@click.argument('file', type=click.File(encoding='utf-8'))
@click.pass_context
def schedule_tasks(ctx, columns, delimiter, file):
    """Schedule tasks from a CSV input file.

    The following columns are expected, and can be set through the -c option:

     - type: the type of the task to be scheduled (mandatory)

     - args: the arguments passed to the task (JSON list, defaults to an empty
       list)

     - kwargs: the keyword arguments passed to the task (JSON object, defaults
       to an empty dict)

     - next_run: the date at which the task should run (datetime, defaults to
       now)

    The CSV can be read either from a named file, or from stdin (use - as
    filename).

    Use sample:

    cat scheduling-task.txt | \
        python3 -m swh.scheduler.cli \
            --database 'service=swh-scheduler-dev' \
            task schedule \
                --columns type --columns kwargs --columns policy \
                --delimiter ';' -

    """
    tasks = []
    now = arrow.utcnow()
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    reader = csv.reader(file, delimiter=delimiter)
    for line in reader:
        task = dict(zip(columns, line))
        args = json.loads(task.pop('args', '[]'))
        kwargs = json.loads(task.pop('kwargs', '{}'))
        task['arguments'] = {
            'args': args,
            'kwargs': kwargs,
        }
        task['next_run'] = DATETIME.convert(task.get('next_run', now),
                                            None, None)
        tasks.append(task)

    created = scheduler.create_tasks(tasks)

    output = [
        'Created %d tasks\n' % len(created),
    ]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo_via_pager('\n'.join(output))


@task.command('add')
@click.argument('type', nargs=1, required=True)
@click.argument('options', nargs=-1)
@click.option('--policy', '-p', default='recurring',
              type=click.Choice(['recurring', 'oneshot']))
@click.option('--priority', '-P', default=None,
              type=click.Choice(['low', 'normal', 'high']))
@click.option('--next-run', '-n', default=None)
@click.pass_context
def schedule_task(ctx, type, options, policy, priority, next_run):
    """Schedule one task from arguments.

    The first argument is the name of the task type, further ones are
    positional and keyword argument(s) of the task, in YAML format.
    Keyword args are of the form key=value.

    Usage sample:

    swh-scheduler --database 'service=swh-scheduler' \
        task add swh-lister-pypi

    swh-scheduler --database 'service=swh-scheduler' \
        task add swh-lister-debian --policy=oneshot distribution=stretch

    Note: if the priority is not given, the task won't have the priority set,
    which is considered as the lowest priority level.
    """
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    now = arrow.utcnow()

    (args, kw) = parse_options(options)
    task = {'type': type,
            'policy': policy,
            'priority': priority,
            'arguments': {
                'args': args,
                'kwargs': kw,
                },
            'next_run': DATETIME.convert(next_run or now,
                                         None, None),
            }
    created = scheduler.create_tasks([task])

    output = [
        'Created %d tasks\n' % len(created),
    ]
    for task in created:
        output.append(pretty_print_task(task))

    click.echo('\n'.join(output))


@task.command('schedule_origins')
@click.argument('type', nargs=1, required=True)
@click.argument('options', nargs=-1)
@click.option('--batch-size', '-b', 'origin_batch_size',
              default=10, show_default=True, type=int,
              help="Number of origins per task")
@click.option('--min-id',
              default=0, show_default=True, type=int,
              help="Only schedule tasks for origins whose ID is greater")
@click.option('--max-id',
              default=None, type=int,
              help="Only schedule tasks for origins whose ID is lower")
@click.option('--storage-url', '-g',
              help="URL of the (graph) storage API")
@click.option('--dry-run/--no-dry-run', is_flag=True,
              default=False,
              help='List only what would be scheduled.')
@click.pass_context
def schedule_origin_metadata_index(
        ctx, type, options, storage_url, origin_batch_size,
        min_id, max_id, dry_run):
    """Schedules tasks for origins that are already known.

    The first argument is the name of the task type, further ones are
    keyword argument(s) of the task in the form key=value, where value is
    in YAML format.

    Usage sample:

    swh-scheduler --database 'service=swh-scheduler' \
        task schedule_origins indexer_origin_metadata
    """
    scheduler = ctx.obj['scheduler']
    storage = get_storage('remote', {'url': storage_url})
    if dry_run:
        scheduler = None

    (args, kw) = parse_options(options)
    if args:
        raise click.ClickException('Only keywords arguments are allowed.')

    origins = iter_origins(storage, origin_from=min_id, origin_to=max_id)
    origin_ids = (origin['id'] for origin in origins)

    schedule_origin_batches(
        scheduler, type, origin_ids, origin_batch_size, kw)


@task.command('list-pending')
@click.argument('task-types', required=True, nargs=-1)
@click.option('--limit', '-l', required=False, type=click.INT,
              help='The maximum number of tasks to fetch')
@click.option('--before', '-b', required=False, type=DATETIME,
              help='List all jobs supposed to run before the given date')
@click.pass_context
def list_pending_tasks(ctx, task_types, limit, before):
    """List the tasks that are going to be run.

    You can override the number of tasks to fetch

    """
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    num_tasks, num_tasks_priority = compute_nb_tasks_from(limit)

    output = []
    for task_type in task_types:
        pending = scheduler.peek_ready_tasks(
            task_type, timestamp=before,
            num_tasks=num_tasks, num_tasks_priority=num_tasks_priority)
        output.append('Found %d %s tasks\n' % (
            len(pending), task_type))

        for task in pending:
            output.append(pretty_print_task(task))

    click.echo('\n'.join(output))


@task.command('list')
@click.option('--task-id', '-i', default=None, multiple=True, metavar='ID',
              help='List only tasks whose id is ID.')
@click.option('--task-type', '-t', default=None, multiple=True, metavar='TYPE',
              help='List only tasks of type TYPE')
@click.option('--limit', '-l', required=False, type=click.INT,
              help='The maximum number of tasks to fetch.')
@click.option('--status', '-s', multiple=True, metavar='STATUS',
              default=None,
              help='List tasks whose status is STATUS.')
@click.option('--policy', '-p', default=None,
              type=click.Choice(['recurring', 'oneshot']),
              help='List tasks whose policy is POLICY.')
@click.option('--priority', '-P', default=None, multiple=True,
              type=click.Choice(['all', 'low', 'normal', 'high']),
              help='List tasks whose priority is PRIORITY.')
@click.option('--before', '-b', required=False, type=DATETIME,
              metavar='DATETIME',
              help='Limit to tasks supposed to run before the given date.')
@click.option('--after', '-a', required=False, type=DATETIME,
              metavar='DATETIME',
              help='Limit to tasks supposed to run after the given date.')
@click.option('--list-runs', '-r', is_flag=True, default=False,
              help='Also list past executions of each task.')
@click.pass_context
def list_tasks(ctx, task_id, task_type, limit, status, policy, priority,
               before, after, list_runs):
    """List tasks.
    """
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    if not task_type:
        task_type = [x['type'] for x in scheduler.get_task_types()]

    # if task_id is not given, default value for status is
    #  'next_run_not_scheduled'
    # if task_id is given, default status is 'all'
    if task_id is None and status is None:
        status = ['next_run_not_scheduled']
    if status and 'all' in status:
        status = None

    if priority and 'all' in priority:
        priority = None

    output = []
    tasks = scheduler.search_tasks(
        task_id=task_id,
        task_type=task_type,
        status=status, priority=priority, policy=policy,
        before=before, after=after,
        limit=limit)
    if list_runs:
        runs = {t['id']: [] for t in tasks}
        for r in scheduler.get_task_runs([task['id'] for task in tasks]):
            runs[r['task']].append(r)
    else:
        runs = {}

    output.append('Found %d tasks\n' % (
        len(tasks)))
    for task in tasks:
        output.append(pretty_print_task(task, full=True))
        if runs.get(task['id']):
            output.append(click.style('  Executions:', bold=True))
            for run in runs[task['id']]:
                output.append(pretty_print_run(run, indent=4))

    click.echo('\n'.join(output))


@task.command('respawn')
@click.argument('task-ids', required=True, nargs=-1)
@click.option('--next-run', '-n', required=False, type=DATETIME,
              metavar='DATETIME', default=None,
              help='Re spawn the selected tasks at this date')
@click.pass_context
def respawn_tasks(ctx, task_ids, next_run):
    """Respawn tasks.

    Respawn tasks given by their ids (see the 'task list' command to
    find task ids) at the given date (immediately by default).

    Eg.

       swh-scheduler task respawn 1 3 12
    """
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')
    if next_run is None:
        next_run = arrow.utcnow()
    output = []

    scheduler.set_status_tasks(
        task_ids, status='next_run_not_scheduled', next_run=next_run)
    output.append('Respawn tasks %s\n' % (task_ids,))

    click.echo('\n'.join(output))


@task.command('archive')
@click.option('--before', '-b', default=None,
              help='''Task whose ended date is anterior will be archived.
                      Default to current month's first day.''')
@click.option('--after', '-a', default=None,
              help='''Task whose ended date is after the specified date will
                      be archived. Default to prior month's first day.''')
@click.option('--batch-index', default=1000, type=click.INT,
              help='Batch size of tasks to read from db to archive')
@click.option('--bulk-index', default=200, type=click.INT,
              help='Batch size of tasks to bulk index')
@click.option('--batch-clean', default=1000, type=click.INT,
              help='Batch size of task to clean after archival')
@click.option('--dry-run/--no-dry-run', is_flag=True, default=False,
              help='Default to list only what would be archived.')
@click.option('--verbose', is_flag=True, default=False,
              help='Verbose mode')
@click.option('--cleanup/--no-cleanup', is_flag=True, default=True,
              help='Clean up archived tasks (default)')
@click.option('--start-from', type=click.INT, default=-1,
              help='(Optional) default task id to start from. Default is -1.')
@click.pass_context
def archive_tasks(ctx, before, after, batch_index, bulk_index, batch_clean,
                  dry_run, verbose, cleanup, start_from):
    """Archive task/task_run whose (task_type is 'oneshot' and task_status
       is 'completed') or (task_type is 'recurring' and task_status is
       'disabled').

       With --dry-run flag set (default), only list those.

    """
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    es_client = SWHElasticSearchClient()
    logging.basicConfig(level=logging.DEBUG if verbose else logging.INFO)
    log = logging.getLogger('swh.scheduler.cli.archive')
    logging.getLogger('urllib3').setLevel(logging.WARN)
    logging.getLogger('elasticsearch').setLevel(logging.WARN)
    if dry_run:
        log.info('**DRY-RUN** (only reading db)')
    if not cleanup:
        log.info('**NO CLEANUP**')

    now = arrow.utcnow()

    # Default to archive tasks from a rolling month starting the week
    # prior to the current one
    if not before:
        before = now.shift(weeks=-1).format('YYYY-MM-DD')

    if not after:
        after = now.shift(weeks=-1).shift(months=-1).format('YYYY-MM-DD')

    log.debug('index: %s; cleanup: %s; period: [%s ; %s]' % (
        not dry_run, not dry_run and cleanup, after, before))

    def group_by_index_name(data, es_client=es_client):
        """Given a data record, determine the index's name through its ending
           date. This varies greatly depending on the task_run's
           status.

        """
        date = data.get('started')
        if not date:
            date = data['scheduled']
        return es_client.compute_index_name(date.year, date.month)

    def index_data(before, last_id, batch_index):
        tasks_in = scheduler.filter_task_to_archive(
            after, before, last_id=last_id, limit=batch_index)
        for index_name, tasks_group in itertools.groupby(
                tasks_in, key=group_by_index_name):
            log.debug('Index tasks to %s' % index_name)
            if dry_run:
                for task in tasks_group:
                    yield task
                continue

            yield from es_client.streaming_bulk(
                index_name, tasks_group, source=['task_id', 'task_run_id'],
                chunk_size=bulk_index, log=log)

    gen = index_data(before, last_id=start_from, batch_index=batch_index)
    if cleanup:
        for task_ids in utils.grouper(gen, n=batch_clean):
            task_ids = list(task_ids)
            log.info('Clean up %s tasks: [%s, ...]' % (
                len(task_ids), task_ids[0]))
            if dry_run:  # no clean up
                continue
            ctx.obj['scheduler'].delete_archived_tasks(task_ids)
    else:
        for task_ids in utils.grouper(gen, n=batch_index):
            task_ids = list(task_ids)
            log.info('Indexed %s tasks: [%s, ...]' % (
                len(task_ids), task_ids[0]))


@cli.command('runner')
@click.option('--period', '-p', default=0,
              help=('Period (in s) at witch pending tasks are checked and '
                    'executed. Set to 0 (default) for a one shot.'))
@click.pass_context
def runner(ctx, period):
    """Starts a swh-scheduler runner service.

    This process is responsible for checking for ready-to-run tasks and
    schedule them."""
    from swh.scheduler.celery_backend.runner import run_ready_tasks
    from swh.scheduler.celery_backend.config import build_app

    app = build_app(ctx.obj['config'].get('celery'))
    app.set_current()

    logger = logging.getLogger(__name__ + '.runner')
    scheduler = ctx.obj['scheduler']
    logger.debug('Scheduler %s' % scheduler)
    try:
        while True:
            logger.debug('Run ready tasks')
            try:
                ntasks = len(run_ready_tasks(scheduler, app))
                if ntasks:
                    logger.info('Scheduled %s tasks', ntasks)
            except Exception:
                logger.exception('Unexpected error in run_ready_tasks()')
            if not period:
                break
            time.sleep(period)
    except KeyboardInterrupt:
        ctx.exit(0)


@cli.command('listener')
@click.pass_context
def listener(ctx):
    """Starts a swh-scheduler listener service.

    This service is responsible for listening at task lifecycle events and
    handle their workflow status in the database."""
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')

    from swh.scheduler.celery_backend.config import build_app
    app = build_app(ctx.obj['config'].get('celery'))
    app.set_current()

    from swh.scheduler.celery_backend.listener import event_monitor
    event_monitor(app, backend=scheduler)


@cli.command('api-server')
@click.option('--host', default='0.0.0.0',
              help="Host to run the scheduler server api")
@click.option('--port', default=5008, type=click.INT,
              help="Binding port of the server")
@click.option('--debug/--nodebug', default=None,
              help=("Indicates if the server should run in debug mode. "
                    "Defaults to True if log-level is DEBUG, False otherwise.")
              )
@click.pass_context
def api_server(ctx, host, port, debug):
    """Starts a swh-scheduler API HTTP server.
    """
    if ctx.obj['config']['scheduler']['cls'] == 'remote':
        click.echo("The API server can only be started with a 'local' "
                   "configuration", err=True)
        ctx.exit(1)

    from swh.scheduler.api import server
    server.app.config.update(ctx.obj['config'])
    if debug is None:
        debug = ctx.obj['loglevel'] <= logging.DEBUG
    server.app.run(host, port=port, debug=bool(debug))


@cli.group('task-type')
@click.pass_context
def task_type(ctx):
    """Manipulate task types."""
    pass


@task_type.command('list')
@click.option('--verbose', '-v', is_flag=True, default=False,
              help='Verbose mode')
@click.option('--task_type', '-t', multiple=True, default=None,
              help='List task types of given type')
@click.option('--task_name', '-n', multiple=True, default=None,
              help='List task types of given backend task name')
@click.pass_context
def list_task_types(ctx, verbose, task_type, task_name):
    click.echo("Known task types:")
    if verbose:
        tmpl = click.style('{type}: ', bold=True) + '''{backend_name}
  {description}
  interval: {default_interval} [{min_interval}, {max_interval}]
  backoff_factor: {backoff_factor}
  max_queue_length: {max_queue_length}
  num_retries: {num_retries}
  retry_delay: {retry_delay}
'''
    else:
        tmpl = '{type}:\n  {description}'
    for tasktype in sorted(ctx.obj['scheduler'].get_task_types(),
                           key=lambda x: x['type']):
        if task_type and tasktype['type'] not in task_type:
            continue
        if task_name and tasktype['backend_name'] not in task_name:
            continue
        click.echo(tmpl.format(**tasktype))


@task_type.command('add')
@click.argument('type', required=1)
@click.argument('task-name', required=1)
@click.argument('description', required=1)
@click.option('--default-interval', '-i', default='90 days',
              help='Default interval ("90 days" by default)')
@click.option('--min-interval', default=None,
              help='Minimum interval (default interval if not set)')
@click.option('--max-interval', '-i', default=None,
              help='Maximal interval (default interval if not set)')
@click.option('--backoff-factor', '-f', type=float, default=1,
              help='Backoff factor')
@click.pass_context
def add_task_type(ctx, type, task_name, description,
                  default_interval, min_interval, max_interval,
                  backoff_factor):
    """Create a new task type
    """
    scheduler = ctx.obj['scheduler']
    if not scheduler:
        raise ValueError('Scheduler class (local/remote) must be instantiated')
    task_type = dict(
        type=type,
        backend_name=task_name,
        description=description,
        default_interval=default_interval,
        min_interval=min_interval,
        max_interval=max_interval,
        backoff_factor=backoff_factor,
        max_queue_length=None,
        num_retries=None,
        retry_delay=None,
        )
    scheduler.create_task_type(task_type)
    click.echo('OK')


@cli.command('updater')
@click.option('--verbose/--no-verbose', '-v', default=False,
              help='Verbose mode')
@click.pass_context
def updater(ctx, verbose):
    """Insert tasks in the scheduler from the scheduler-updater's events

    """
    from swh.scheduler.updater.writer import UpdaterWriter
    UpdaterWriter(**ctx.obj['config']).run()


@cli.command('ghtorrent')
@click.option('--verbose/--no-verbose', '-v', default=False,
              help='Verbose mode')
@click.pass_context
def ghtorrent(ctx, verbose):
    """Consume events from ghtorrent and write them to cache.

    """
    from swh.scheduler.updater.ghtorrent import GHTorrentConsumer
    from swh.scheduler.updater.backend import SchedulerUpdaterBackend

    ght_config = ctx.obj['config'].get('ghtorrent', {})
    back_config = ctx.obj['config'].get('scheduler_updater', {})
    backend = SchedulerUpdaterBackend(**back_config)
    GHTorrentConsumer(backend, **ght_config).run()


def main():
    return cli(auto_envvar_prefix='SWH_SCHEDULER')


if __name__ == '__main__':
    main()



from time import sleep
from celery.result import GroupResult
from celery.result import AsyncResult

import pytest

from swh.scheduler.utils import create_task_dict
from swh.scheduler.celery_backend.runner import run_ready_tasks


def test_ping(swh_app, celery_session_worker):
    res = swh_app.send_task(
        'swh.scheduler.tests.tasks.ping')
    assert res
    res.wait()
    assert res.successful()
    assert res.result == 'OK'


def test_multiping(swh_app, celery_session_worker):
    "Test that a task that spawns subtasks (group) works"
    res = swh_app.send_task(
        'swh.scheduler.tests.tasks.multiping', n=5)
    assert res

    res.wait()
    assert res.successful()

    # retrieve the GroupResult for this task and wait for all the subtasks
    # to complete
    promise_id = res.result
    assert promise_id
    promise = GroupResult.restore(promise_id, app=swh_app)
    for i in range(5):
        if promise.ready():
            break
        sleep(1)

    results = [x.get() for x in promise.results]
    for i in range(5):
        assert ("OK (kw={'i': %s})" % i) in results


@pytest.mark.db
def test_scheduler_fixture(swh_app, celery_session_worker, swh_scheduler):
    "Test that the scheduler fixture works properly"
    task_type = swh_scheduler.get_task_type('swh-test-ping')

    assert task_type
    assert task_type['backend_name'] == 'swh.scheduler.tests.tasks.ping'

    swh_scheduler.create_tasks([create_task_dict(
        'swh-test-ping', 'oneshot')])

    backend_tasks = run_ready_tasks(swh_scheduler, swh_app)
    assert backend_tasks
    for task in backend_tasks:
        # Make sure the task completed
        AsyncResult(id=task['backend_id']).get()


@pytest.mark.db
def test_task_return_value(swh_app, celery_session_worker, swh_scheduler):
    task_type = swh_scheduler.get_task_type('swh-test-add')
    assert task_type
    assert task_type['backend_name'] == 'swh.scheduler.tests.tasks.add'

    swh_scheduler.create_tasks([create_task_dict(
        'swh-test-add', 'oneshot', 12, 30)])

    backend_tasks = run_ready_tasks(swh_scheduler, swh_app)
    assert len(backend_tasks) == 1
    task = backend_tasks[0]
    value = AsyncResult(id=task['backend_id']).get()
    assert value == 42


@pytest.mark.db
def test_task_exception(swh_app, celery_session_worker, swh_scheduler):
    task_type = swh_scheduler.get_task_type('swh-test-error')
    assert task_type
    assert task_type['backend_name'] == 'swh.scheduler.tests.tasks.error'

    swh_scheduler.create_tasks([create_task_dict(
        'swh-test-error', 'oneshot')])

    backend_tasks = run_ready_tasks(swh_scheduler, swh_app)
    assert len(backend_tasks) == 1

    task = backend_tasks[0]
    result = AsyncResult(id=task['backend_id'])
    with pytest.raises(NotImplementedError):
        result.get()



# Copyright (C) 2015-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging

from arrow import Arrow, utcnow
import psycopg2.pool
import psycopg2.extras
from psycopg2.extensions import AsIs

from swh.core.db import BaseDb
from swh.core.db.common import db_transaction, db_transaction_generator


logger = logging.getLogger(__name__)


def adapt_arrow(arrow):
    return AsIs("'%s'::timestamptz" % arrow.isoformat())


psycopg2.extensions.register_adapter(dict, psycopg2.extras.Json)
psycopg2.extensions.register_adapter(Arrow, adapt_arrow)


def format_query(query, keys):
    """Format a query with the given keys"""

    query_keys = ', '.join(keys)
    placeholders = ', '.join(['%s'] * len(keys))

    return query.format(keys=query_keys, placeholders=placeholders)


class SchedulerBackend:
    """Backend for the Software Heritage scheduling database.

    """

    def __init__(self, db, min_pool_conns=1, max_pool_conns=10):
        """
        Args:
            db_conn: either a libpq connection string, or a psycopg2 connection

        """
        if isinstance(db, psycopg2.extensions.connection):
            self._pool = None
            self._db = BaseDb(db)
        else:
            self._pool = psycopg2.pool.ThreadedConnectionPool(
                min_pool_conns, max_pool_conns, db,
                cursor_factory=psycopg2.extras.RealDictCursor,
            )
            self._db = None

    def get_db(self):
        if self._db:
            return self._db
        return BaseDb.from_pool(self._pool)

    task_type_keys = [
        'type', 'description', 'backend_name', 'default_interval',
        'min_interval', 'max_interval', 'backoff_factor', 'max_queue_length',
        'num_retries', 'retry_delay',
    ]

    @db_transaction()
    def create_task_type(self, task_type, db=None, cur=None):
        """Create a new task type ready for scheduling.

        Args:
            task_type (dict): a dictionary with the following keys:

                - type (str): an identifier for the task type
                - description (str): a human-readable description of what the
                  task does
                - backend_name (str): the name of the task in the
                  job-scheduling backend
                - default_interval (datetime.timedelta): the default interval
                  between two task runs
                - min_interval (datetime.timedelta): the minimum interval
                  between two task runs
                - max_interval (datetime.timedelta): the maximum interval
                  between two task runs
                - backoff_factor (float): the factor by which the interval
                  changes at each run
                - max_queue_length (int): the maximum length of the task queue
                  for this task type

        """
        keys = [key for key in self.task_type_keys if key in task_type]
        query = format_query(
            """insert into task_type ({keys}) values ({placeholders})""",
            keys)
        cur.execute(query, [task_type[key] for key in keys])

    @db_transaction()
    def get_task_type(self, task_type_name, db=None, cur=None):
        """Retrieve the task type with id task_type_name"""
        query = format_query(
            "select {keys} from task_type where type=%s",
            self.task_type_keys,
        )
        cur.execute(query, (task_type_name,))
        return cur.fetchone()

    @db_transaction()
    def get_task_types(self, db=None, cur=None):
        """Retrieve all registered task types"""
        query = format_query(
            "select {keys} from task_type",
            self.task_type_keys,
        )
        cur.execute(query)
        return cur.fetchall()

    task_create_keys = [
        'type', 'arguments', 'next_run', 'policy', 'status', 'retries_left',
        'priority'
    ]
    task_keys = task_create_keys + ['id', 'current_interval', 'status']

    @db_transaction()
    def create_tasks(self, tasks, policy='recurring', db=None, cur=None):
        """Create new tasks.

        Args:
            tasks (list): each task is a dictionary with the following keys:

                - type (str): the task type
                - arguments (dict): the arguments for the task runner, keys:

                      - args (list of str): arguments
                      - kwargs (dict str -> str): keyword arguments

                - next_run (datetime.datetime): the next scheduled run for the
                  task

        Returns:
            a list of created tasks.

        """
        cur.execute('select swh_scheduler_mktemp_task()')
        db.copy_to(tasks, 'tmp_task', self.task_create_keys,
                   default_values={
                       'policy': policy,
                       'status': 'next_run_not_scheduled'
                   },
                   cur=cur)
        query = format_query(
            'select {keys} from swh_scheduler_create_tasks_from_temp()',
            self.task_keys,
        )
        cur.execute(query)
        return cur.fetchall()

    @db_transaction()
    def set_status_tasks(self, task_ids, status='disabled', next_run=None,
                         db=None, cur=None):
        """Set the tasks' status whose ids are listed.

        If given, also set the next_run date.
        """
        if not task_ids:
            return
        query = ["UPDATE task SET status = %s"]
        args = [status]
        if next_run:
            query.append(', next_run = %s')
            args.append(next_run)
        query.append(" WHERE id IN %s")
        args.append(tuple(task_ids))

        cur.execute(''.join(query), args)

    @db_transaction()
    def disable_tasks(self, task_ids, db=None, cur=None):
        """Disable the tasks whose ids are listed."""
        return self.set_status_tasks(task_ids, db=db, cur=cur)

    @db_transaction()
    def search_tasks(self, task_id=None, task_type=None, status=None,
                     priority=None, policy=None, before=None, after=None,
                     limit=None, db=None, cur=None):
        """Search tasks from selected criterions"""
        where = []
        args = []

        if task_id:
            if isinstance(task_id, (str, int)):
                where.append('id = %s')
            else:
                where.append('id in %s')
                task_id = tuple(task_id)
            args.append(task_id)
        if task_type:
            if isinstance(task_type, str):
                where.append('type = %s')
            else:
                where.append('type in %s')
                task_type = tuple(task_type)
            args.append(task_type)
        if status:
            if isinstance(status, str):
                where.append('status = %s')
            else:
                where.append('status in %s')
                status = tuple(status)
            args.append(status)
        if priority:
            if isinstance(priority, str):
                where.append('priority = %s')
            else:
                priority = tuple(priority)
                where.append('priority in %s')
            args.append(priority)
        if policy:
            where.append('policy = %s')
            args.append(policy)
        if before:
            where.append('next_run <= %s')
            args.append(before)
        if after:
            where.append('next_run >= %s')
            args.append(after)

        query = 'select * from task'
        if where:
            query += ' where ' + ' and '.join(where)
        if limit:
            query += ' limit %s :: bigint'
            args.append(limit)
        cur.execute(query, args)
        return cur.fetchall()

    @db_transaction()
    def get_tasks(self, task_ids, db=None, cur=None):
        """Retrieve the info of tasks whose ids are listed."""
        query = format_query('select {keys} from task where id in %s',
                             self.task_keys)
        cur.execute(query, (tuple(task_ids),))
        return cur.fetchall()

    @db_transaction()
    def peek_ready_tasks(self, task_type, timestamp=None, num_tasks=None,
                         num_tasks_priority=None,
                         db=None, cur=None):
        """Fetch the list of ready tasks

        Args:
            task_type (str): filtering task per their type
            timestamp (datetime.datetime): peek tasks that need to be executed
                before that timestamp
            num_tasks (int): only peek at num_tasks tasks (with no priority)
            num_tasks_priority (int): only peek at num_tasks_priority
                                      tasks (with priority)

        Returns:
            a list of tasks

        """
        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            '''select * from swh_scheduler_peek_ready_tasks(
                %s, %s, %s :: bigint, %s :: bigint)''',
            (task_type, timestamp, num_tasks, num_tasks_priority)
        )
        logger.debug('PEEK %s => %s' % (task_type, cur.rowcount))
        return cur.fetchall()

    @db_transaction()
    def grab_ready_tasks(self, task_type, timestamp=None, num_tasks=None,
                         num_tasks_priority=None, db=None, cur=None):
        """Fetch the list of ready tasks, and mark them as scheduled

        Args:
            task_type (str): filtering task per their type
            timestamp (datetime.datetime): grab tasks that need to be executed
                before that timestamp
            num_tasks (int): only grab num_tasks tasks (with no priority)
            num_tasks_priority (int): only grab oneshot num_tasks tasks (with
                                      priorities)

        Returns:
            a list of tasks

        """
        if timestamp is None:
            timestamp = utcnow()
        cur.execute(
            '''select * from swh_scheduler_grab_ready_tasks(
                 %s, %s, %s :: bigint, %s :: bigint)''',
            (task_type, timestamp, num_tasks, num_tasks_priority)
        )
        logger.debug('GRAB %s => %s' % (task_type, cur.rowcount))
        return cur.fetchall()

    task_run_create_keys = ['task', 'backend_id', 'scheduled', 'metadata']

    @db_transaction()
    def schedule_task_run(self, task_id, backend_id, metadata=None,
                          timestamp=None, db=None, cur=None):
        """Mark a given task as scheduled, adding a task_run entry in the database.

        Args:
            task_id (int): the identifier for the task being scheduled
            backend_id (str): the identifier of the job in the backend
            metadata (dict): metadata to add to the task_run entry
            timestamp (datetime.datetime): the instant the event occurred

        Returns:
            a fresh task_run entry

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            'select * from swh_scheduler_schedule_task_run(%s, %s, %s, %s)',
            (task_id, backend_id, metadata, timestamp)
        )

        return cur.fetchone()

    @db_transaction()
    def mass_schedule_task_runs(self, task_runs, db=None, cur=None):
        """Schedule a bunch of task runs.

        Args:
            task_runs (list): a list of dicts with keys:

                - task (int): the identifier for the task being scheduled
                - backend_id (str): the identifier of the job in the backend
                - metadata (dict): metadata to add to the task_run entry
                - scheduled (datetime.datetime): the instant the event occurred

        Returns:
            None
        """
        cur.execute('select swh_scheduler_mktemp_task_run()')
        db.copy_to(task_runs, 'tmp_task_run', self.task_run_create_keys,
                   cur=cur)
        cur.execute('select swh_scheduler_schedule_task_run_from_temp()')

    @db_transaction()
    def start_task_run(self, backend_id, metadata=None, timestamp=None,
                       db=None, cur=None):
        """Mark a given task as started, updating the corresponding task_run
           entry in the database.

        Args:
            backend_id (str): the identifier of the job in the backend
            metadata (dict): metadata to add to the task_run entry
            timestamp (datetime.datetime): the instant the event occurred

        Returns:
            the updated task_run entry

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            'select * from swh_scheduler_start_task_run(%s, %s, %s)',
            (backend_id, metadata, timestamp)
        )

        return cur.fetchone()

    @db_transaction()
    def end_task_run(self, backend_id, status, metadata=None, timestamp=None,
                     result=None, db=None, cur=None):
        """Mark a given task as ended, updating the corresponding task_run entry in the
        database.

        Args:
            backend_id (str): the identifier of the job in the backend
            status (str): how the task ended; one of: 'eventful', 'uneventful',
                'failed'
            metadata (dict): metadata to add to the task_run entry
            timestamp (datetime.datetime): the instant the event occurred

        Returns:
            the updated task_run entry

        """

        if metadata is None:
            metadata = {}

        if timestamp is None:
            timestamp = utcnow()

        cur.execute(
            'select * from swh_scheduler_end_task_run(%s, %s, %s, %s)',
            (backend_id, status, metadata, timestamp)
        )
        return cur.fetchone()

    @db_transaction_generator()
    def filter_task_to_archive(self, after_ts, before_ts, limit=10, last_id=-1,
                               db=None, cur=None):
        """Returns the list of task/task_run prior to a given date to archive.

        """
        last_task_run_id = None
        while True:
            row = None
            cur.execute(
                "select * from swh_scheduler_task_to_archive(%s, %s, %s, %s)",
                (after_ts, before_ts, last_id, limit)
            )
            for row in cur:
                # nested type index does not accept bare values
                # transform it as a dict to comply with this
                row['arguments']['args'] = {
                    i: v for i, v in enumerate(row['arguments']['args'])
                }
                kwargs = row['arguments']['kwargs']
                row['arguments']['kwargs'] = json.dumps(kwargs)
                yield row

            if not row:
                break
            _id = row.get('task_id')
            _task_run_id = row.get('task_run_id')
            if last_id == _id and last_task_run_id == _task_run_id:
                break
            last_id = _id
            last_task_run_id = _task_run_id

    @db_transaction()
    def delete_archived_tasks(self, task_ids, db=None, cur=None):
        """Delete archived tasks as much as possible. Only the task_ids whose
           complete associated task_run have been cleaned up will be.

        """
        _task_ids = _task_run_ids = []
        for task_id in task_ids:
            _task_ids.append(task_id['task_id'])
            _task_run_ids.append(task_id['task_run_id'])

        cur.execute(
            "select * from swh_scheduler_delete_archived_tasks(%s, %s)",
            (_task_ids, _task_run_ids))

    task_run_keys = ['id', 'task', 'backend_id', 'scheduled',
                     'started', 'ended', 'metadata', 'status', ]

    @db_transaction()
    def get_task_runs(self, task_ids, limit=None, db=None, cur=None):
        """Search task run for a task id"""
        where = []
        args = []

        if task_ids:
            if isinstance(task_ids, (str, int)):
                where.append('task = %s')
            else:
                where.append('task in %s')
                task_ids = tuple(task_ids)
            args.append(task_ids)
        else:
            return ()

        query = 'select * from task_run where ' + ' and '.join(where)
        if limit:
            query += ' limit %s :: bigint'
            args.append(limit)
        cur.execute(query, args)
        return cur.fetchall()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import click
import logging
import time

from arrow import utcnow

from swh.core import utils
from swh.scheduler import get_scheduler
from swh.scheduler.utils import create_oneshot_task_dict
from swh.scheduler.updater.backend import SchedulerUpdaterBackend


class UpdaterWriter:
    """Updater writer in charge of updating the scheduler db with latest
       prioritized oneshot tasks

       In effect, this:
       - reads the events from scheduler updater's db
       - converts those events into priority oneshot tasks
       - dumps them into the scheduler db

    """

    def __init__(self, **config):
        self.config = config
        if self.config['scheduler_updater']['cls'] != 'local':
            raise ValueError(
                'The scheduler_updater can only be a cls=local for now')
        self.scheduler_updater_backend = SchedulerUpdaterBackend(
            **self.config['scheduler_updater']['args'])
        self.scheduler_backend = get_scheduler(**self.config['scheduler'])
        self.pause = self.config.get('updater_writer', {}).get('pause', 10)
        self.log = logging.getLogger(
            'swh.scheduler.updater.writer.UpdaterWriter')

    def convert_to_oneshot_task(self, event):
        """Given an event, convert it into oneshot task with priority

        Args:
            event (dict): The event to convert to task

        """
        if event['origin_type'] == 'git':
            return create_oneshot_task_dict(
                'origin-update-git',
                event['url'],
                priority='normal')
        self.log.warning('Type %s is not supported for now, only git' % (
            event['origin_type'], ))
        return None

    def write_event_to_scheduler(self, events):
        """Write events to the scheduler and yield ids when done"""
        # convert events to oneshot tasks
        oneshot_tasks = filter(lambda e: e is not None,
                               map(self.convert_to_oneshot_task, events))
        # write event to scheduler
        self.scheduler_backend.create_tasks(list(oneshot_tasks))
        for e in events:
            yield e['url']

    def run(self):
        """First retrieve events from cache (including origin_type, cnt),
           then convert them into oneshot tasks with priority, then
           write them to the scheduler db, at last remove them from
           cache.

        """
        while True:
            timestamp = utcnow()
            events = list(self.scheduler_updater_backend.cache_read(timestamp))
            if not events:
                break
            for urls in utils.grouper(self.write_event_to_scheduler(events),
                                      n=100):
                self.scheduler_updater_backend.cache_remove(urls)
            time.sleep(self.pause)


@click.command()
@click.option('--verbose/--no-verbose', '-v', default=False,
              help='Verbose mode')
@click.pass_context
def main(ctx, verbose):
    click.echo("Deprecated! Use 'swh-scheduler updater' instead.",
               err=True)
    ctx.exit(1)


if __name__ == '__main__':
    main()



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from hypothesis import given
from hypothesis.strategies import sets

from swh.scheduler.updater.events import SWHEvent

from . import from_regex


@given(urls=sets(
    from_regex(
        r'^https://somewhere[.]org/[a-z0-9]{5,7}/[a-z0-9]{3,10}$'),
    min_size=10, max_size=15))
def test_cache_read(urls, swh_scheduler_updater):
    # beware that the fixture is only called once for all the tests
    # generated by hypothesis, so the db is not cleared between calls.
    # see the end of
    #   https://hypothesis.works/articles/hypothesis-pytest-fixtures/
    def gen_events(urls):
        for url in urls:
            yield SWHEvent({
                'url': url,
                'type': 'create',
                'origin_type': 'git',
            })
    known_urls = set(e['url'] for e in
                     swh_scheduler_updater.cache_read(limit=1000000))
    swh_scheduler_updater.cache_put(gen_events(urls))
    new_urls = {u.strip() for u in urls} - known_urls
    all_urls = set(e['url'] for e in
                   swh_scheduler_updater.cache_read(limit=1000000))
    assert (all_urls - known_urls) == new_urls



import pytest
import glob
import os

from swh.core.utils import numfile_sortkey as sortkey
from swh.scheduler.updater.backend import SchedulerUpdaterBackend
from swh.scheduler.tests import SQL_DIR
import swh.scheduler.tests.conftest  # noqa


DUMP_FILES = os.path.join(SQL_DIR, 'updater', '*.sql')


@pytest.fixture
def swh_scheduler_updater(request, postgresql_proc, postgresql):
    config = {
        'db': 'postgresql://{user}@{host}:{port}/{dbname}'.format(
            host=postgresql_proc.host,
            port=postgresql_proc.port,
            user='postgres',
            dbname='tests')
    }

    all_dump_files = sorted(glob.glob(DUMP_FILES), key=sortkey)

    cursor = postgresql.cursor()
    for fname in all_dump_files:
        with open(fname) as fobj:
            cursor.execute(fobj.read())
    postgresql.commit()

    backend = SchedulerUpdaterBackend(**config)
    return backend



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from arrow import utcnow

try:
    from hypothesis.strategies import from_regex
except ImportError:
    from hypothesis.strategies import text

    # Revert to using basic text generation
    def from_regex(*args, **kwargs):
        return text()


class UpdaterTestUtil:
    """Mixin intended for event generation purposes

    """
    def _make_event(self, event_type, name, origin_type):
        return {
            'type': event_type,
            'repo': {
                'name': name,
            },
            'created_at': utcnow(),
            'origin_type': origin_type,
        }

    def _make_events(self, events):
        for event_type, repo_name, origin_type in events:
            yield self._make_event(event_type, repo_name, origin_type)

    def _make_incomplete_event(self, event_type, name, origin_type,
                               missing_data_key):
        event = self._make_event(event_type, name, origin_type)
        del event[missing_data_key]
        return event

    def _make_incomplete_events(self, events):
        for event_type, repo_name, origin_type, missing_data_key in events:
            yield self._make_incomplete_event(event_type, repo_name,
                                              origin_type, missing_data_key)

    def _make_simple_event(self, event_type, name, origin_type):
        return {
            'type': event_type,
            'url': 'https://fakeurl/%s' % name,
            'origin_type': origin_type,
            'created_at': utcnow(),
        }



from celery import group, current_app as app


@app.task(name='swh.scheduler.tests.tasks.ping', bind=True)
def ping(self, **kw):
    # check this is a SWHTask
    assert hasattr(self, 'log')
    assert not hasattr(self, 'run_task')
    assert 'SWHTask' in [x.__name__ for x in self.__class__.__mro__]
    self.log.debug(self.name)
    if kw:
        return 'OK (kw=%s)' % kw
    return 'OK'


@app.task(name='swh.scheduler.tests.tasks.multiping', bind=True)
def multiping(self, n=10):
    promise = group(ping.s(i=i) for i in range(n))()
    self.log.debug('%s OK (spawned %s subtasks)' % (self.name, n))
    promise.save()
    return promise.id


@app.task(name='swh.scheduler.tests.tasks.error')
def not_implemented():
    raise NotImplementedError('Nope')


@app.task(name='swh.scheduler.tests.tasks.add')
def add(x, y):
    return x + y



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import logging

from flask import request, Flask

from swh.core import config
from swh.core.api import (decode_request,
                          error_handler,
                          encode_data_server as encode_data)

from swh.core.api import negotiate, JSONFormatter, MsgpackFormatter
from swh.scheduler import get_scheduler as get_scheduler_from


app = Flask(__name__)
scheduler = None


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


def get_sched():
    global scheduler
    if not scheduler:
        scheduler = get_scheduler_from(**app.config['scheduler'])
    return scheduler


def has_no_empty_params(rule):
    return len(rule.defaults or ()) >= len(rule.arguments or ())


@app.route('/')
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def index():
    return 'SWH Scheduler API server'


@app.route('/close_connection', methods=['GET', 'POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def close_connection():
    return get_sched().close_connection()


@app.route('/set_status_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def set_status_tasks():
    return get_sched().set_status_tasks(**decode_request(request))


@app.route('/create_task_type', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def create_task_type():
    return get_sched().create_task_type(**decode_request(request))


@app.route('/get_task_type', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def get_task_type():
    return get_sched().get_task_type(**decode_request(request))


@app.route('/get_task_types', methods=['GET', 'POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def get_task_types():
    return get_sched().get_task_types(**decode_request(request))


@app.route('/create_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def create_tasks():
    return get_sched().create_tasks(**decode_request(request))


@app.route('/disable_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def disable_tasks():
    return get_sched().disable_tasks(**decode_request(request))


@app.route('/get_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def get_tasks():
    return get_sched().get_tasks(**decode_request(request))


@app.route('/get_task_runs', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def get_task_runs():
    return get_sched().get_task_runs(**decode_request(request))


@app.route('/search_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def search_tasks():
    return get_sched().search_tasks(**decode_request(request))


@app.route('/peek_ready_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def peek_ready_tasks():
    return get_sched().peek_ready_tasks(**decode_request(request))


@app.route('/grab_ready_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def grab_ready_tasks():
    return get_sched().grab_ready_tasks(**decode_request(request))


@app.route('/schedule_task_run', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def schedule_task_run():
    return get_sched().schedule_task_run(**decode_request(request))


@app.route('/mass_schedule_task_runs', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def mass_schedule_task_runs():
    return get_sched().mass_schedule_task_runs(**decode_request(request))


@app.route('/start_task_run', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def start_task_run():
    return get_sched().start_task_run(**decode_request(request))


@app.route('/end_task_run', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def end_task_run():
    return get_sched().end_task_run(**decode_request(request))


@app.route('/filter_task_to_archive', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def filter_task_to_archive():
    return get_sched().filter_task_to_archive(**decode_request(request))


@app.route('/delete_archived_tasks', methods=['POST'])
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def delete_archived_tasks():
    return get_sched().delete_archived_tasks(**decode_request(request))


@app.route("/site-map")
@negotiate(MsgpackFormatter)
@negotiate(JSONFormatter)
def site_map():
    links = []
    sched = get_sched()
    for rule in app.url_map.iter_rules():
        if has_no_empty_params(rule) and hasattr(sched, rule.endpoint):
            links.append(dict(
                rule=rule.rule,
                description=getattr(sched, rule.endpoint).__doc__))
    # links is now a list of url, endpoint tuples
    return links


def load_and_check_config(config_file, type='local'):
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_file (str): Path to the configuration file to load
        type (str): configuration type. For 'local' type, more
                    checks are done.

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_file:
        raise EnvironmentError('Configuration file must be defined')

    if not os.path.exists(config_file):
        raise FileNotFoundError('Configuration file %s does not exist' % (
            config_file, ))

    cfg = config.read(config_file)

    vcfg = cfg.get('scheduler')
    if not vcfg:
        raise KeyError("Missing '%scheduler' configuration")

    if type == 'local':
        cls = vcfg.get('cls')
        if cls != 'local':
            raise ValueError(
                "The scheduler backend can only be started with a 'local' "
                "configuration")

        args = vcfg.get('args')
        if not args:
            raise KeyError(
                "Invalid configuration; missing 'args' config entry")

        db = args.get('db')
        if not db:
            raise KeyError(
                "Invalid configuration; missing 'db' config entry")

    return cfg


api_cfg = None


def make_app_from_configfile():
    """Run the WSGI app from the webserver, loading the configuration from
       a configuration file.

       SWH_CONFIG_FILENAME environment variable defines the
       configuration path to load.

    """
    global api_cfg
    if not api_cfg:
        config_file = os.environ.get('SWH_CONFIG_FILENAME')
        api_cfg = load_and_check_config(config_file)
        app.config.update(api_cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app


if __name__ == '__main__':
    print('Please use the "swh-scheduler api-server" command')



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from arrow import utcnow
import psycopg2.pool
import psycopg2.extras

from swh.core.db import BaseDb
from swh.core.db.common import db_transaction, db_transaction_generator
from swh.scheduler.backend import format_query


class SchedulerUpdaterBackend:
    CONFIG_BASE_FILENAME = 'backend/scheduler-updater'
#        'cache_read_limit': ('int', 1000),

    def __init__(self, db, cache_read_limit=1000,
                 min_pool_conns=1, max_pool_conns=10):
        """
        Args:
            db_conn: either a libpq connection string, or a psycopg2 connection

        """
        if isinstance(db, psycopg2.extensions.connection):
            self._pool = None
            self._db = BaseDb(db)
        else:
            self._pool = psycopg2.pool.ThreadedConnectionPool(
                min_pool_conns, max_pool_conns, db,
                cursor_factory=psycopg2.extras.RealDictCursor,
            )
            self._db = None
        self.limit = cache_read_limit

    def get_db(self):
        if self._db:
            return self._db
        return BaseDb.from_pool(self._pool)

    cache_put_keys = ['url', 'cnt', 'last_seen', 'origin_type']

    @db_transaction()
    def cache_put(self, events, timestamp=None, db=None, cur=None):
        """Write new events in the backend.

        """
        cur.execute('select swh_mktemp_cache()')
        db.copy_to(prepare_events(events, timestamp),
                   'tmp_cache', self.cache_put_keys, cur=cur)
        cur.execute('select swh_cache_put()')

    cache_read_keys = ['id', 'url', 'origin_type', 'cnt', 'first_seen',
                       'last_seen']

    @db_transaction_generator()
    def cache_read(self, timestamp=None, limit=None, db=None, cur=None):
        """Read events from the cache prior to timestamp.

        Note that limit=None does not mean 'no limit' but use the default
        limit (see cache_read_limit constructor argument).

        """
        if not timestamp:
            timestamp = utcnow()

        if not limit:
            limit = self.limit

        q = format_query('select {keys} from swh_cache_read(%s, %s)',
                         self.cache_read_keys)
        cur.execute(q, (timestamp, limit))
        yield from cur.fetchall()

    @db_transaction()
    def cache_remove(self, entries, db=None, cur=None):
        """Clean events from the cache

        """
        q = 'delete from cache where url in (%s)' % (
            ', '.join(("'%s'" % e for e in entries)), )
        cur.execute(q)


def prepare_events(events, timestamp=None):
    if timestamp is None:
        timestamp = utcnow()
    outevents = []
    urls = []
    for e in events:
        event = e.get()
        url = event['url'].strip()
        if event['last_seen'] is None:
            event['last_seen'] = timestamp
        event['url'] = url

        if url in urls:
            idx = urls.index(url)
            urls.append(urls.pop(idx))
            pevent = outevents.pop(idx)
            event['cnt'] += pevent['cnt']
            event['last_seen'] = max(
                event['last_seen'], pevent['last_seen'])
        else:
            urls.append(url)
        outevents.append(event)
    return outevents



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.core.api import SWHRemoteAPI


class RemoteScheduler(SWHRemoteAPI):
    """Proxy to a remote scheduler API

    """
    def close_connection(self):
        return self.post('close_connection', {})

    def set_status_tasks(self, task_ids, status='disabled', next_run=None):
        return self.post('set_status_tasks', dict(
            task_ids=task_ids, status=status, next_run=next_run))

    def create_task_type(self, task_type):
        return self.post('create_task_type', {'task_type': task_type})

    def get_task_type(self, task_type_name):
        return self.post('get_task_type', {'task_type_name': task_type_name})

    def get_task_types(self):
        return self.post('get_task_types', {})

    def create_tasks(self, tasks):
        return self.post('create_tasks', {'tasks': tasks})

    def disable_tasks(self, task_ids):
        return self.post('disable_tasks', {'task_ids': task_ids})

    def get_tasks(self, task_ids):
        return self.post('get_tasks', {'task_ids': task_ids})

    def get_task_runs(self, task_ids, limit=None):
        return self.post(
            'get_task_runs', {'task_ids': task_ids, 'limit': limit})

    def search_tasks(self, task_id=None, task_type=None, status=None,
                     priority=None, policy=None, before=None, after=None,
                     limit=None):
        return self.post('search_tasks', dict(
            task_id=task_id, task_type=task_type, status=status,
            priority=priority, policy=policy, before=before, after=after,
            limit=limit))

    def peek_ready_tasks(self, task_type, timestamp=None, num_tasks=None,
                         num_tasks_priority=None):
        return self.post('peek_ready_tasks', {
            'task_type': task_type,
            'timestamp': timestamp,
            'num_tasks': num_tasks,
            'num_tasks_priority': num_tasks_priority,
        })

    def grab_ready_tasks(self, task_type, timestamp=None, num_tasks=None,
                         num_tasks_priority=None):
        return self.post('grab_ready_tasks', {
            'task_type': task_type,
            'timestamp': timestamp,
            'num_tasks': num_tasks,
            'num_tasks_priority': num_tasks_priority,
        })

    def schedule_task_run(self, task_id, backend_id, metadata=None,
                          timestamp=None):
        return self.post('schedule_task_run', {
            'task_id': task_id,
            'backend_id': backend_id,
            'metadata': metadata,
            'timestamp': timestamp,
        })

    def mass_schedule_task_runs(self, task_runs):
        return self.post('mass_schedule_task_runs', {'task_runs': task_runs})

    def start_task_run(self, backend_id, metadata=None, timestamp=None):
        return self.post('start_task_run', {
            'backend_id': backend_id,
            'metadata': metadata,
            'timestamp': timestamp,
        })

    def end_task_run(self, backend_id, status, metadata=None, timestamp=None):
        return self.post('end_task_run', {
            'backend_id': backend_id,
            'status': status,
            'metadata': metadata,
            'timestamp': timestamp,
        })

    def filter_task_to_archive(self, after_ts, before_ts, limit=10,
                               last_id=-1):
        return self.post('filter_task_to_archive', {
            'after_ts': after_ts,
            'before_ts': before_ts,
            'limit': limit,
            'last_id': last_id,
        })

    def delete_archived_tasks(self, task_ids):
        return self.post('delete_archived_tasks', {'task_ids': task_ids})



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json

from kombu import Connection, Exchange, Queue
from kombu.common import collect_replies

from swh.core.config import merge_configs

from swh.scheduler.updater.events import SWHEvent
from swh.scheduler.updater.consumer import UpdaterConsumer
from swh.scheduler.updater.backend import SchedulerUpdaterBackend


events = {
    # ghtorrent events related to github events (interesting)
    'evt': [
        'commitcomment', 'create', 'delete', 'deployment',
        'deploymentstatus', 'download', 'follow', 'fork', 'forkapply',
        'gist', 'gollum', 'issuecomment', 'issues', 'member',
        'membership', 'pagebuild', 'public', 'pullrequest',
        'pullrequestreviewcomment', 'push', 'release', 'repository',
        'status', 'teamadd', 'watch'
    ],
    # ghtorrent events related to mongodb insert (not interesting)
    'ent': [
        'commit_comments', 'commits', 'followers', 'forks',
        'geo_cache', 'issue_comments', 'issue_events', 'issues',
        'org_members', 'pull_request_comments', 'pull_requests',
        'repo_collaborators', 'repo_labels', 'repos', 'users', 'watchers'
    ]
}

INTERESTING_EVENT_KEYS = ['type', 'repo', 'created_at']

DEFAULT_CONFIG = {
    'ghtorrent': {
        'batch_cache_write': 1000,
        'rabbitmq': {
            'prefetch_read': 100,
            'conn': {
                'url': 'amqp://guest:guest@localhost:5672',
                'exchange_name': 'ght-streams',
                'routing_key': 'something',
                'queue_name': 'fake-events',
            },
        },
    },
    'scheduler_updater': {
        'cls': 'local',
        'args': {
            'db': 'dbname=softwareheritage-scheduler-updater-dev',
            'cache_read_limit': 1000,
        },
    },
}


class GHTorrentConsumer(UpdaterConsumer):
    """GHTorrent events consumer

    """
    connection_class = Connection

    def __init__(self, **config):
        self.config = merge_configs(DEFAULT_CONFIG, config)

        ght_config = self.config['ghtorrent']
        rmq_config = ght_config['rabbitmq']
        self.prefetch_read = int(rmq_config.get('prefetch_read', 100))

        exchange = Exchange(
            rmq_config['conn']['exchange_name'],
            'topic', durable=True)
        routing_key = rmq_config['conn']['routing_key']
        self.queue = Queue(rmq_config['conn']['queue_name'],
                           exchange=exchange,
                           routing_key=routing_key,
                           auto_delete=True)

        if self.config['scheduler_updater']['cls'] != 'local':
            raise ValueError(
                'The scheduler_updater can only be a cls=local for now')
        backend = SchedulerUpdaterBackend(
            **self.config['scheduler_updater']['args'])

        super().__init__(backend, ght_config.get('batch_cache_write', 1000))

    def has_events(self):
        """Always has events

        """
        return True

    def convert_event(self, event):
        """Given ghtorrent event, convert it to a SWHEvent instance.

        """
        if isinstance(event, str):
            event = json.loads(event)
        for k in INTERESTING_EVENT_KEYS:
            if k not in event:
                if hasattr(self, 'log'):
                    self.log.warning(
                        'Event should have the \'%s\' entry defined' % k)
                return None

        _type = event['type'].lower().rstrip('Event')
        _repo_name = 'https://github.com/%s' % event['repo']['name']

        return SWHEvent({
            'type': _type,
            'url': _repo_name,
            'last_seen': event['created_at'],
            'origin_type': 'git',
        })

    def open_connection(self):
        """Open rabbitmq connection

        """
        self.conn = self.connection_class(
            self.config['ghtorrent']['rabbitmq']['conn']['url'])
        self.conn.connect()
        self.channel = self.conn.channel()

    def close_connection(self):
        """Close rabbitmq connection

        """
        self.channel.close()
        self.conn.release()

    def consume_events(self):
        """Consume and yield queue messages

        """
        yield from collect_replies(
            self.conn, self.channel, self.queue,
            no_ack=False, limit=self.prefetch_read)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
import requests

from swh.core.tests.server_testing import ServerTestFixture
from swh.scheduler import get_scheduler
from swh.scheduler.api.server import app
from swh.scheduler.tests.test_scheduler import CommonSchedulerTest


class RemoteSchedulerTest(CommonSchedulerTest, ServerTestFixture,
                          unittest.TestCase):
    """Test the remote scheduler API.

    This class doesn't define any tests as we want identical
    functionality between local and remote scheduler. All the tests are
    therefore defined in CommonSchedulerTest.
    """

    def setUp(self):
        self.config = {
            'scheduler': {
                'cls': 'local',
                'args': {
                    'db': 'dbname=%s' % self.TEST_DB_NAME,
                }
            }
        }
        self.app = app  # this will setup the local scheduler...
        super().setUp()
        # accessible through a remote scheduler accessible on the
        # given port
        self.backend = get_scheduler('remote', {'url': self.url()})

    def test_site_map(self):
        sitemap = requests.get(self.url() + 'site-map')
        assert sitemap.headers['Content-Type'] == 'application/json'
        sitemap = sitemap.json()

        rules = set(x['rule'] for x in sitemap)
        # we expect at least these rules
        expected_rules = set('/'+rule for rule in (
            'set_status_tasks', 'create_task_type',
            'get_task_type', 'get_task_types', 'create_tasks', 'disable_tasks',
            'get_tasks', 'search_tasks', 'peek_ready_tasks',
            'grab_ready_tasks', 'schedule_task_run', 'mass_schedule_task_runs',
            'start_task_run', 'end_task_run', 'filter_task_to_archive',
            'delete_archived_tasks'))
        assert rules.issuperset(expected_rules), expected_rules - rules



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
import unittest
from glob import glob

import pytest

from swh.core.utils import numfile_sortkey as sortkey
from swh.core.tests.db_testing import DbTestFixture
from swh.scheduler.tests import SQL_DIR
from swh.scheduler.updater.events import LISTENED_EVENTS, SWHEvent
from swh.scheduler.updater.writer import UpdaterWriter

from . import UpdaterTestUtil


@pytest.mark.db
class CommonSchedulerTest(DbTestFixture):
    TEST_SCHED_DB = 'softwareheritage-scheduler-test'
    TEST_SCHED_DUMP = os.path.join(SQL_DIR, '*.sql')

    TEST_SCHED_UPDATER_DB = 'softwareheritage-scheduler-updater-test'
    TEST_SCHED_UPDATER_DUMP = os.path.join(SQL_DIR, 'updater', '*.sql')

    @classmethod
    def setUpClass(cls):
        cls.add_db(cls.TEST_SCHED_DB,
                   [(sqlfn, 'psql') for sqlfn in
                    sorted(glob(cls.TEST_SCHED_DUMP), key=sortkey)])
        cls.add_db(cls.TEST_SCHED_UPDATER_DB,
                   [(sqlfn, 'psql') for sqlfn in
                    sorted(glob(cls.TEST_SCHED_UPDATER_DUMP), key=sortkey)])
        super().setUpClass()

    def tearDown(self):
        self.reset_db_tables(self.TEST_SCHED_UPDATER_DB)
        self.reset_db_tables(self.TEST_SCHED_DB,
                             excluded=['task_type', 'priority_ratio'])
        super().tearDown()


class UpdaterWriterTest(UpdaterTestUtil, CommonSchedulerTest,
                        unittest.TestCase):
    def setUp(self):
        super().setUp()

        config = {
            'scheduler': {
                'cls': 'local',
                'args': {
                    'db': 'dbname=softwareheritage-scheduler-test',
                },
            },
            'scheduler_updater': {
                'cls': 'local',
                'args': {
                    'db':
                    'dbname=softwareheritage-scheduler-updater-test',
                    'cache_read_limit': 5,
                },
            },
            'updater_writer': {
                'pause': 0.1,
                'verbose': False,
            },
        }
        self.writer = UpdaterWriter(**config)
        self.scheduler_backend = self.writer.scheduler_backend
        self.scheduler_updater_backend = self.writer.scheduler_updater_backend

    def test_run_ko(self):
        """Only git tasks are supported for now, other types are dismissed.

        """
        ready_events = [
            SWHEvent(
                self._make_simple_event(event_type, 'origin-%s' % i,
                                        'svn'))
            for i, event_type in enumerate(LISTENED_EVENTS)
        ]

        expected_length = len(ready_events)

        self.scheduler_updater_backend.cache_put(ready_events)
        data = list(self.scheduler_updater_backend.cache_read())
        self.assertEqual(len(data), expected_length)

        r = self.scheduler_backend.peek_ready_tasks(
            'origin-update-git')

        # first read on an empty scheduling db results with nothing in it
        self.assertEqual(len(r), 0)

        # Read from cache to scheduler db
        self.writer.run()

        r = self.scheduler_backend.peek_ready_tasks(
            'origin-update-git')

        # other reads after writes are still empty since it's not supported
        self.assertEqual(len(r), 0)

    def test_run_ok(self):
        """Only git origin are supported for now

        """
        ready_events = [
            SWHEvent(
                self._make_simple_event(event_type, 'origin-%s' % i, 'git'))
            for i, event_type in enumerate(LISTENED_EVENTS)
        ]

        expected_length = len(ready_events)

        self.scheduler_updater_backend.cache_put(ready_events)

        data = list(self.scheduler_updater_backend.cache_read())
        self.assertEqual(len(data), expected_length)

        r = self.scheduler_backend.peek_ready_tasks(
            'origin-update-git')

        # first read on an empty scheduling db results with nothing in it
        self.assertEqual(len(r), 0)

        # Read from cache to scheduler db
        self.writer.run()

        # now, we should have scheduling task ready
        r = self.scheduler_backend.peek_ready_tasks(
            'origin-update-git')

        self.assertEqual(len(r), expected_length)

        # Check the task has been scheduled
        for t in r:
            self.assertEqual(t['type'], 'origin-update-git')
            self.assertEqual(t['priority'], 'normal')
            self.assertEqual(t['policy'], 'oneshot')
            self.assertEqual(t['status'], 'next_run_not_scheduled')

        # writer has nothing to do now
        self.writer.run()

        # so no more data in cache
        data = list(self.scheduler_updater_backend.cache_read())

        self.assertEqual(len(data), 0)

        # provided, no runner is ran, still the same amount of scheduling tasks
        r = self.scheduler_backend.peek_ready_tasks(
            'origin-update-git')

        self.assertEqual(len(r), expected_length)



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from itertools import chain

from hypothesis import given, settings, HealthCheck
from hypothesis.strategies import lists, sampled_from, text, tuples

from swh.scheduler.updater.consumer import UpdaterConsumer
from swh.scheduler.updater.events import LISTENED_EVENTS, SWHEvent

from . import UpdaterTestUtil, from_regex


class FakeSchedulerUpdaterBackend:
    def __init__(self):
        self.events = []

    def cache_put(self, events):
        self.events.append(events)


class FakeUpdaterConsumerBase(UpdaterConsumer):
    def __init__(self, backend):
        super().__init__(backend)
        self.connection_opened = False
        self.connection_closed = False
        self.consume_called = False
        self.has_events_called = False

    def open_connection(self):
        self.connection_opened = True

    def close_connection(self):
        self.connection_closed = True

    def convert_event(self, event):
        pass


class FakeUpdaterConsumerRaise(FakeUpdaterConsumerBase):
    def has_events(self):
        self.has_events_called = True
        return True

    def consume_events(self):
        self.consume_called = True
        raise ValueError('Broken stuff')


class UpdaterConsumerRaisingTest(unittest.TestCase):
    def setUp(self):
        self.updater = FakeUpdaterConsumerRaise(
            FakeSchedulerUpdaterBackend())

    def test_running_raise(self):
        """Raising during run should finish fine.

        """
        # given
        self.assertEqual(self.updater.count, 0)
        self.assertEqual(self.updater.seen_events, set())
        self.assertEqual(self.updater.events, [])

        # when
        with self.assertRaisesRegex(ValueError, 'Broken stuff'):
            self.updater.run()

        # then
        self.assertEqual(self.updater.count, 0)
        self.assertEqual(self.updater.seen_events, set())
        self.assertEqual(self.updater.events, [])
        self.assertTrue(self.updater.connection_opened)
        self.assertTrue(self.updater.has_events_called)
        self.assertTrue(self.updater.connection_closed)
        self.assertTrue(self.updater.consume_called)


class FakeUpdaterConsumerNoEvent(FakeUpdaterConsumerBase):
    def has_events(self):
        self.has_events_called = True
        return False

    def consume_events(self):
        self.consume_called = True


class UpdaterConsumerNoEventTest(unittest.TestCase):
    def setUp(self):
        self.updater = FakeUpdaterConsumerNoEvent(
            FakeSchedulerUpdaterBackend())

    def test_running_does_not_consume(self):
        """Run with no events should do just fine"""
        # given
        self.assertEqual(self.updater.count, 0)
        self.assertEqual(self.updater.seen_events, set())
        self.assertEqual(self.updater.events, [])

        # when
        self.updater.run()

        # then
        self.assertEqual(self.updater.count, 0)
        self.assertEqual(self.updater.seen_events, set())
        self.assertEqual(self.updater.events, [])
        self.assertTrue(self.updater.connection_opened)
        self.assertTrue(self.updater.has_events_called)
        self.assertTrue(self.updater.connection_closed)
        self.assertFalse(self.updater.consume_called)


EVENT_KEYS = ['type', 'repo', 'created_at', 'origin_type']


class FakeUpdaterConsumer(FakeUpdaterConsumerBase):
    def __init__(self, backend, messages):
        super().__init__(backend)
        self.messages = messages
        self.debug = False

    def has_events(self):
        self.has_events_called = True
        return len(self.messages) > 0

    def consume_events(self):
        self.consume_called = True
        for msg in self.messages:
            yield msg
            self.messages.pop()

    def convert_event(self, event, keys=EVENT_KEYS):
        for k in keys:
            v = event.get(k)
            if v is None:
                return None

        e = {
            'type': event['type'],
            'url': 'https://fake.url/%s' % event['repo']['name'],
            'last_seen': event['created_at'],
            'origin_type': event['origin_type'],
        }
        return SWHEvent(e)


class UpdaterConsumerWithEventTest(UpdaterTestUtil, unittest.TestCase):
    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(lists(tuples(sampled_from(LISTENED_EVENTS),  # event type
                        from_regex(r'^[a-z0-9]{5,10}/[a-z0-9]{7,12}$'),  # name
                        text()),                        # origin type
                 min_size=3, max_size=10),
           lists(tuples(text(),                         # event type
                        from_regex(r'^[a-z0-9]{5,7}/[a-z0-9]{3,10}$'),  # name
                        text()),                        # origin type
                 min_size=3, max_size=10),
           lists(tuples(sampled_from(LISTENED_EVENTS),  # event type
                        from_regex(r'^[a-z0-9]{5,10}/[a-z0-9]{7,12}$'),  # name
                        text(),                     # origin type
                        sampled_from(EVENT_KEYS)),  # keys to drop
                 min_size=3, max_size=10))
    def test_running(self, events, uninteresting_events, incomplete_events):
        """Interesting events are written to cache, others are dropped

        """
        # given
        ready_events = self._make_events(events)
        ready_uninteresting_events = self._make_events(uninteresting_events)
        ready_incomplete_events = self._make_incomplete_events(
            incomplete_events)

        updater = FakeUpdaterConsumer(
            FakeSchedulerUpdaterBackend(),
            list(chain(
                ready_events, ready_incomplete_events,
                ready_uninteresting_events)))

        self.assertEqual(updater.count, 0)
        self.assertEqual(updater.seen_events, set())
        self.assertEqual(updater.events, [])

        # when
        updater.run()

        # then
        self.assertEqual(updater.count, 0)
        self.assertEqual(updater.seen_events, set())
        self.assertEqual(updater.events, [])
        self.assertTrue(updater.connection_opened)
        self.assertTrue(updater.has_events_called)
        self.assertTrue(updater.connection_closed)
        self.assertTrue(updater.consume_called)

        self.assertEqual(updater.messages, [])
        # uninteresting or incomplete events are dropped
        self.assertTrue(len(updater.backend.events), len(events))



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os
import urllib.parse

from celery import Celery
from celery.signals import setup_logging, celeryd_after_setup
from celery.utils.log import ColorFormatter
from celery.worker.control import Panel

from kombu import Exchange, Queue
from kombu.five import monotonic as _monotonic

import requests

from swh.scheduler import CONFIG as SWH_CONFIG

from swh.core.config import load_named_config, merge_configs
from swh.core.logger import JournalHandler

DEFAULT_CONFIG_NAME = 'worker'
CONFIG_NAME_ENVVAR = 'SWH_WORKER_INSTANCE'
CONFIG_NAME_TEMPLATE = 'worker/%s'

DEFAULT_CONFIG = {
    'task_broker': ('str', 'amqp://guest@localhost//'),
    'task_modules': ('list[str]', []),
    'task_queues': ('list[str]', []),
    'task_soft_time_limit': ('int', 0),
}

logger = logging.getLogger(__name__)


@setup_logging.connect
def setup_log_handler(loglevel=None, logfile=None, format=None, colorize=None,
                      log_console=None, log_journal=None, **kwargs):
    """Setup logging according to Software Heritage preferences.

    We use the command-line loglevel for tasks only, as we never
    really care about the debug messages from celery.
    """
    if loglevel is None:
        loglevel = logging.DEBUG
    if isinstance(loglevel, str):
        loglevel = logging._nameToLevel[loglevel]

    formatter = logging.Formatter(format)

    root_logger = logging.getLogger('')
    root_logger.setLevel(logging.INFO)

    log_target = os.environ.get('SWH_LOG_TARGET', 'console')
    if log_target == 'console':
        log_console = True
    elif log_target == 'journal':
        log_journal = True

    # this looks for log levels *higher* than DEBUG
    if loglevel <= logging.DEBUG and log_console is None:
        log_console = True

    if log_console:
        color_formatter = ColorFormatter(format) if colorize else formatter
        console = logging.StreamHandler()
        console.setLevel(logging.DEBUG)
        console.setFormatter(color_formatter)
        root_logger.addHandler(console)

    if log_journal:
        systemd_journal = JournalHandler()
        systemd_journal.setLevel(logging.DEBUG)
        systemd_journal.setFormatter(formatter)
        root_logger.addHandler(systemd_journal)

    logging.getLogger('celery').setLevel(logging.INFO)
    # Silence amqp heartbeat_tick messages
    logger = logging.getLogger('amqp')
    logger.addFilter(lambda record: not record.msg.startswith(
        'heartbeat_tick'))
    logger.setLevel(loglevel)
    # Silence useless "Starting new HTTP connection" messages
    logging.getLogger('urllib3').setLevel(logging.WARNING)

    logging.getLogger('swh').setLevel(loglevel)
    # get_task_logger makes the swh tasks loggers children of celery.task
    logging.getLogger('celery.task').setLevel(loglevel)
    return loglevel


@celeryd_after_setup.connect
def setup_queues_and_tasks(sender, instance, **kwargs):
    """Signal called on worker start.

    This automatically registers swh.scheduler.task.Task subclasses as
    available celery tasks.

    This also subscribes the worker to the "implicit" per-task queues defined
    for these task classes.

    """
    logger.info('Setup Queues & Tasks for %s', sender)
    instance.app.conf['worker_name'] = sender


@Panel.register
def monotonic(state):
    """Get the current value for the monotonic clock"""
    return {'monotonic': _monotonic()}


def route_for_task(name, args, kwargs, options, task=None, **kw):
    """Route tasks according to the task_queue attribute in the task class"""
    if name is not None and name.startswith('swh.'):
        return {'queue': name}


def get_queue_stats(app, queue_name):
    """Get the statistics regarding a queue on the broker.

    Arguments:
      queue_name: name of the queue to check

    Returns a dictionary raw from the RabbitMQ management API;
    or `None` if the current configuration does not use RabbitMQ.

    Interesting keys:
     - Consumers (number of consumers for the queue)
     - messages (number of messages in queue)
     - messages_unacknowledged (number of messages currently being
       processed)

    Documentation: https://www.rabbitmq.com/management.html#http-api
    """

    conn_info = app.connection().info()
    if conn_info['transport'] == 'memory':
        # We're running in a test environment, without RabbitMQ.
        return None
    url = 'http://{hostname}:{port}/api/queues/{vhost}/{queue}'.format(
        hostname=conn_info['hostname'],
        port=conn_info['port'] + 10000,
        vhost=urllib.parse.quote(conn_info['virtual_host'], safe=''),
        queue=urllib.parse.quote(queue_name, safe=''),
    )
    credentials = (conn_info['userid'], conn_info['password'])
    r = requests.get(url, auth=credentials)
    if r.status_code == 404:
        return {}
    if r.status_code != 200:
        raise ValueError('Got error %s when reading queue stats: %s' % (
            r.status_code, r.json()))
    return r.json()


def get_queue_length(app, queue_name):
    """Shortcut to get a queue's length"""
    stats = get_queue_stats(app, queue_name)
    if stats:
        return stats.get('messages')


def register_task_class(app, name, cls):
    """Register a class-based task under the given name"""
    if name in app.tasks:
        return

    task_instance = cls()
    task_instance.name = name
    app.register_task(task_instance)


INSTANCE_NAME = os.environ.get(CONFIG_NAME_ENVVAR)
CONFIG_NAME = os.environ.get('SWH_CONFIG_FILENAME')
CONFIG = {}

if CONFIG_NAME:
    # load the celery config from the main config file given as
    # SWH_CONFIG_FILENAME environment variable.
    # This is expected to have a [celery] section in which we have the
    # celery specific configuration.
    SWH_CONFIG.clear()
    SWH_CONFIG.update(load_named_config(CONFIG_NAME))
    CONFIG = SWH_CONFIG.get('celery')

if not CONFIG:
    # otherwise, back to compat config loading mechanism
    if INSTANCE_NAME:
        CONFIG_NAME = CONFIG_NAME_TEMPLATE % INSTANCE_NAME
    else:
        CONFIG_NAME = DEFAULT_CONFIG_NAME

    # Load the Celery config
    CONFIG = load_named_config(CONFIG_NAME, DEFAULT_CONFIG)

# Celery Queues
CELERY_QUEUES = [Queue('celery', Exchange('celery'), routing_key='celery')]

CELERY_DEFAULT_CONFIG = dict(
    # Timezone configuration: all in UTC
    enable_utc=True,
    timezone='UTC',
    # Imported modules
    imports=CONFIG.get('task_modules', []),
    # Time (in seconds, or a timedelta object) for when after stored task
    # tombstones will be deleted. None means to never expire results.
    result_expires=None,
    # A string identifying the default serialization method to use. Can
    # be json (default), pickle, yaml, msgpack, or any custom
    # serialization methods that have been registered with
    task_serializer='msgpack',
    # Result serialization format
    result_serializer='msgpack',
    # Late ack means the task messages will be acknowledged after the task has
    # been executed, not just before, which is the default behavior.
    task_acks_late=True,
    # A string identifying the default serialization method to use.
    # Can be pickle (default), json, yaml, msgpack or any custom serialization
    # methods that have been registered with kombu.serialization.registry
    accept_content=['msgpack', 'json'],
    # If True the task will report its status as “started”
    # when the task is executed by a worker.
    task_track_started=True,
    # Default compression used for task messages. Can be gzip, bzip2
    # (if available), or any custom compression schemes registered
    # in the Kombu compression registry.
    # result_compression='bzip2',
    # task_compression='bzip2',
    # Disable all rate limits, even if tasks has explicit rate limits set.
    # (Disabling rate limits altogether is recommended if you don’t have any
    # tasks using them.)
    worker_disable_rate_limits=True,
    # Task routing
    task_routes=route_for_task,
    # Allow pool restarts from remote
    worker_pool_restarts=True,
    # Do not prefetch tasks
    worker_prefetch_multiplier=1,
    # Send events
    worker_send_task_events=True,
    # Do not send useless task_sent events
    task_send_sent_event=False,
    )


def build_app(config=None):
    config = merge_configs(
        {k: v for (k, (_, v)) in DEFAULT_CONFIG.items()},
        config or {})

    config['task_queues'] = CELERY_QUEUES + [
        Queue(queue, Exchange(queue), routing_key=queue)
        for queue in config.get('task_queues', ())]
    logger.debug('Creating a Celery app with %s', config)

    # Instantiate the Celery app
    app = Celery(broker=config['task_broker'],
                 task_cls='swh.scheduler.task:SWHTask')
    app.add_defaults(CELERY_DEFAULT_CONFIG)
    app.add_defaults(config)
    return app


app = build_app(CONFIG)

# XXX for BW compat
Celery.get_queue_length = get_queue_length
